from .record import Record

class Study(Record):

    def __init__(self, folder, UID=[], **attributes):
        
        super().__init__(folder, UID, generation=2, **attributes)

    def label(self, row=None):

        if row is None:
            data = self.data()
            if data.empty: return "New Study"
            file = data.index[0]
            descr = data.at[file, 'StudyDescription']
            date = data.at[file, 'StudyDate']
        else:
            descr = row.StudyDescription
            date = row.StudyDate

        label = str(descr)
        label += ' [' + str(date) + ']'
        return label


import os
import shutil
from copy import deepcopy
from datetime import datetime

import pydicom
from pydicom.dataset import Dataset
from pydicom.sequence import Sequence

from .record import Record
from .. import utilities

class Instance(Record):

    def __init__(self, folder, UID=[], **attributes):
        super().__init__(folder, UID, generation=4, **attributes)

    def label(self, row=None):

        if row is None:
            data = self.data()
            if data.empty: return "New Instance"
            file = data.index[0]
            nr = data.at[file, 'InstanceNumber']
        else:
            nr = row.InstanceNumber

        return str(nr).zfill(6)

    @property
    def file(self):
        """Returns the filepath to the instance."""
 
        files = self.files
        if len(files) != 0:  
            return files[0]

    def __getitem__(self, tags):
        """Gets the value of the data elements with specified tags.
        
        Arguments
        ---------
        tags : a string, hexadecimal tuple, or a list of strings and hexadecimal tuples

        Returns
        -------
        A value or a list of values
        """
        in_memory = self.in_memory()
        if not in_memory: 
            self.read()
        values = utilities._read_tags(self.ds, tags)   
        if not in_memory: 
            self.clear()
        return values

    def __setitem__(self, tags, values):
        """
        Sets the value of the data element with given tag.
        """
        on_disk = self.on_disk()
        if on_disk: 
            self.read()
        utilities._set_tags(self.ds, tags, values)
        if on_disk:
            self.write()
            self.clear()

    def write(self):
        """Writes the dataset to disk"""

        if self.on_disk(): return
        ds = self.ds

        # Ensure DICOM hierarchy is respected
        if not 'PatientID' in ds:
            ds.PatientID = pydicom.uid.generate_uid()
        if not 'StudyInstanceUID' in ds:
            ds.StudyInstanceUID = pydicom.uid.generate_uid()
        if not 'SeriesInstanceUID' in ds:
            ds.SeriesInstanceUID = pydicom.uid.generate_uid()    
        if not 'SOPInstanceUID' in ds:
            ds.SOPInstanceUID = pydicom.uid.generate_uid()    

        df = self.data()
        if df.empty: # Data exist in memory only.
            self.folder._append(ds)
        else: 
            file = df.index[0] 
            if not df.loc[file,'created']:  # This is the first change 
                self.folder.dataframe.loc[file,'removed'] = True
                self.folder._append(ds)
            else:   # Update values in dataframe row
                self.folder._update(file, ds)

        self._save_ds()

    def _save_ds(self, file=None):

        if file is None: 
            file = self.file
        try:
            self.ds.save_as(file) 
        except:
            message = "Failed to write to " + file
            message += "\n The file is open in another application, or is being synchronised by a cloud service."
            message += "\n Please close the file or pause the synchronisation and try again."
            self.dialog.information(message)

    def clear(self):
        """Clears the dataset in memory"""

        self.__dict__['ds'] = None 

    def read(self):
        """Reads the dataset into memory."""

        file = self.file
        if file is None: 
            return
        try:
            self.ds = pydicom.dcmread(file)
        except:
            message = "Failed to read " + file
            message += "\n Please read the DICOM folder again via File->Read."
            self.dialog.information(message)            
        return self.ds # return self instead so read can be piped: series.read().copy_to(parent)

    def _copy_to_OBSOLETE(self, ancestor): # Replaced by copy_to in record
        """copy instance to a new ancestor.
        
        dicom_object: Root, Patient, Study, or Series
            If the object is not a series, the missing 
            intermediate generations are automatically created.
        """
        # Generate new instance
        copy = self.__class__(self.folder, UID=ancestor.UID)

        if self.in_memory(): # Create the copy in memory
            copy.__dict__['ds'] = deepcopy(self.ds)
            copy._initialize(self.ds)
            if ancestor.in_memory():
                ancestor.ds.append(copy)
        else: # Create copy on disk
            self.read()
            copy.__dict__['ds'] = self.ds
            copy._initialize(self.ds)
            copy.folder._append(copy.ds)
            copy._save_ds()
            self.clear()
            if ancestor.in_memory():
                ancestor.ds.append(copy) 
            else:       
                copy.clear()
            
        return copy

    def save_OBSOLETE(self):
        """
        Saves all changes made in the instance
        """
        self.write() 
        rows = self.folder.dataframe.SOPInstanceUID == self.UID[-1]
        data = self.folder.dataframe[rows] 
        created = data.created[data.created]
        removed = data.removed[data.removed]

        if data.shape[0] == 2: # instance has been modified
            created = created.index
            removed = removed.index
            if created.empty or removed.empty: 
                message = 'DICOM dataset ' + self.file
                message += "\n Source data have been corrupted."
                message += "\n In your DICOM folder, remove the folder Weasel"
                message += "\n and delete the .csv file."
                message += "\n Then read the DICOM folder again."
                self.dialog.error(message)
                return
            try: # save changes in original file
                shutil.copyfile(created[0], removed[0])
            except:
                message = "DICOM files have been removed or are open in another application."
                message += '\n Close the files and try again.'
                self.dialog.error(message)
                return
            os.remove(created[0])
            self.folder.dataframe.drop(created, inplace=True)
            self.folder.dataframe.loc[removed, 'removed'] = False

        elif not removed.empty: # instance has been deleted
            index = removed.index
            os.remove(index[0])
            self.folder.dataframe.drop(index, inplace=True)
            
        elif not created.empty: # instance has been newly created
            self.folder.dataframe.loc[created.index, 'created'] = False

    def restore_OBSOLETE(self):
        """
        Reverses all changes made since the last save.
        """
        in_memory = self.in_memory()
        self.clear() 
        rows = self.folder.dataframe[self.key[-1]] == self.UID[-1]
        data = self.folder.dataframe[rows] 
        created = data.created[data.created]
        removed = data.removed[data.removed]
        if not removed.empty: # restore deleted files
            index = removed.index
            self.folder.dataframe.loc[index, 'removed'] = False
        if not created.empty: # delete new files
            index = created.index
            file = index[0]
            self.folder.dataframe.drop(index, inplace = True)
            if os.path.exists(file): os.remove(file)
        if in_memory: self.read()

    def export(self, path):
        """Export instances to an external folder.

        This will create another copy of the same instance.
        The instance itself will not be removed from the DICOM folder.
        Instead a copy of the file will be copied to the external folder.
        
        Arguments
        ---------
        path : str
            path to an external folder.
        """
        in_memory = self.in_memory()
        if not in_memory: self.read()
        filename = os.path.basename(self.file)
        destination = os.path.join(path, filename)
        self._save_ds(destination)
        if not in_memory: self.clear()

    def _initialize(self, ref_ds=None):
        """Initialize the attributes relevant for the Images"""

        self.ds = utilities._initialize(self.ds, UID=self.UID, ref=ref_ds)
        return

        # overwrite UIDs
        self.ds.PatientID = self.UID[0]
        self.ds.StudyInstanceUID = self.UID[1]
        self.ds.SeriesInstanceUID = self.UID[2]
        self.ds.SOPInstanceUID = self.UID[3]

        # Date and Time of Creation
        dt = datetime.now()
        timeStr = dt.strftime('%H%M%S')  # long format with micro seconds

        self.ds.ContentDate = dt.strftime('%Y%m%d')
        self.ds.ContentTime = timeStr
        self.ds.AcquisitionDate = dt.strftime('%Y%m%d')
        self.ds.AcquisitionTime = timeStr
        self.ds.SeriesDate = dt.strftime('%Y%m%d')
        self.ds.SeriesTime = timeStr
        self.ds.InstanceCreationDate = dt.strftime('%Y%m%d')
        self.ds.InstanceCreationTime = timeStr

        if ref_ds is None: return

        # Series, Instance and Class for Reference
        refd_instance = Dataset()
        refd_instance.ReferencedSOPClassUID = ref_ds.SOPClassUID
        refd_instance.ReferencedSOPInstanceUID = ref_ds.SOPInstanceUID
        refd_instance_sequence = Sequence()
        refd_instance_sequence.append(refd_instance)

        refd_series = Dataset()
        refd_series.ReferencedInstanceSequence = refd_instance_sequence
        refd_series.SeriesInstanceUID = self.ds.SeriesInstanceUID
        refd_series_sequence = Sequence()
        refd_series_sequence.append(refd_series)

        self.ds.ReferencedSeriesSequence = refd_series_sequence


import matplotlib.pyplot as plt
import pandas as pd
import numpy as np


from .folder import Folder

test_folder = 'C:\\Users\\steve\\Dropbox\\Data\\WeaselDevSmallBrain'
test_folder2 = 'C:\\Users\\steve\\Dropbox\\Data\\WeaselDevSmallKid'
test_export_folder = 'C:\\Users\\steve\\Dropbox\\Data\\Tmp'


def test_df():

    columns = ['Bool', 'Name', 'Age', 'Stream', 'Percentage']
    record = {
        'Bool' : [True, True, False, True, False, True] , 
        'Name': ['Ankit', 'Amit', 'Aishwarya', 'Priyanka', 'Priya', 'Shaurya' ],
        'Age': [21, 19, 20, 18, 17, 21],
        'Stream': ['Math', 'Commerce', 'Science', 'Math', 'Math', 'Science'],
        'Percentage': [88, 92, 95, 70, 65, 78] } 
    df = pd.DataFrame(record, 
        columns = columns, 
        index = ['A','A','C',None,'E','F'])
    print(df)

def test_set_array_mip():
    pass

def test_set_slice_array_invert():

    folder = Folder(test_folder2).open() # enhancement: open automatically on init
    instance = folder.series(12).instances(0)
    array = instance.array()
    instance.set_array(-array)
    array_invert = instance.array()

    print('These should be equal: ', np.sum(np.square(array)), np.sum(np.square(array_invert)))
    print('This should be zero: ', np.sum(np.square(array + array_invert)))

    col = 3
    fig = plt.figure(figsize=(16,16))
    i=0
    fig.add_subplot(1,col,i+1)
    plt.imshow(array)
    plt.colorbar()
    i=1
    fig.add_subplot(1,col,i+1)
    plt.imshow(array_invert)
    plt.colorbar()
    i=2
    fig.add_subplot(1,col,i+1)
    plt.imshow(array+array_invert)
    plt.colorbar()
    plt.show()
    folder.restore()

def test_set_array_invert():

    folder = Folder(test_folder2).open() # enhancement: open automatically on init
    series = folder.series(12) # T1-MOLLI multi-TI
    array, _ = series.array()
    invert = series.copy().set_array(-array)
    array_invert, _ = invert.array()
    print('Round-off error (%): ', 100*np.sum(np.square(array + array_invert))/np.sum(np.square(array)))
    col = 3
    fig = plt.figure(figsize=(16,16))
    i=0
    fig.add_subplot(1,col,i+1)
    plt.imshow(array[0,:,:])
    plt.colorbar()
    i=1
    fig.add_subplot(1,col,i+1)
    plt.imshow(array_invert[0,:,:])
    plt.colorbar()
    i=2
    fig.add_subplot(1,col,i+1)
    plt.imshow(array[0,:,:]+array_invert[0,:,:])
    plt.colorbar()
    plt.show()

    folder.restore()

def test_sort_series():

    folder = Folder(test_folder2).open()
    series = folder.series(12) # T1-MOLLI multi-TI
#    data = series.sort(['SliceLocation','InversionTime', 'PatientName'])
    data = series.dataset(['SliceLocation','InversionTime', 'PatientName'])
    print(data[0,0,0,0].__class__.__name__)
    print(data.shape)
    print(data[0,0,0,0].SliceLocation)
    print(data[0,0,0,0].InversionTime)
    print(data[0,0,0,0].PatientName)
    loc = [data[z,0,0,0].SliceLocation for z in range(data.shape[0])]
    print(loc)

def test_read_series_array():

    folder = Folder(test_folder2).open()
    series = folder.series(12) # T1-MOLLI multi-TI
    array, data = series.array(['SliceLocation','InversionTime', 'PatientName'], pixels_first=True)
    print(array.shape)
    print(data.shape)
    print([data[0,0,0,0].SliceLocation, data[0,0,0,0].InversionTime, data[0,0,0,0].PatientName])

def test_read_and_open():

    folder = Folder(path = test_folder)
    folder.scan()
    folder.print()
    folder.open()
    folder.print()

def test_retrieve():

    folder = Folder(path = test_folder) 
    folder.scan()

    print('Printing SeriesDescription of all series in the folder')
    for series in folder.series():
        print(series.SeriesDescription)
    print('Printing Rows for first series in the folder')
    for instance in folder.series(0).instances():
        print(instance.Rows)
    print('Printing PatientName of all patients in the folder')
    for patient in folder.patients():
        print(patient.PatientName)
    print('ID of the first patient in the folder')
    print(folder.patients(0).PatientID)
    print('Description of the second series for the first patient')
    print(folder.patients(0).series(1).SeriesDescription)
    print('Description of the first series of the first study of the first patient')
    print(folder.patients(0).studies(0).series(0).SeriesDescription)

def test_find():

    folder = Folder(path = test_folder) 
    folder.scan()
    print('Find series with series description ax 15 flip and given PatientID')
    series = folder.patients(0).series(SeriesDescription="ax 15 flip", PatientID='RIDER Neuro MRI-5244517593')
    for s in series: print(s.SeriesDescription) 

def test_read_item_instance():

    folder = Folder(path = test_folder) 
    folder.scan()
    tags = ['SeriesDescription', (0x0010, 0x0020), (0x0010, 0x0020), 'PatientID', (0x0011, 0x0020)]
    instance = folder.instances(0)
    print(instance.PatientID)
    print(instance[tags])

def test_read_item():

    folder = Folder(path = test_folder)  
    folder.scan()
    tags = ['SeriesDescription', (0x0010, 0x0020), (0x0010, 0x0020), 'PatientID', (0x0011, 0x0020)]
    series = folder.series(0)
    print(series.PatientID)
    print(series[tags])
    patient = folder.patients(0)
    print(patient[tags])

def test_set_attr_instance():

    folder = Folder(path = test_folder) 
    folder.scan()
    instance = folder.instances(0)

    slice_loc = instance.SliceLocation
    print('Original Slice location: ' + str(slice_loc))
    instance.SliceLocation = slice_loc + 100
    slice_loc = instance.SliceLocation
    print('New Slice location: ' + str(slice_loc))

    acq_time = instance.AcquisitionTime
    print('Original Acquisition Time: ' + str(acq_time))
    instance.AcquisitionTime = '00.00.00'
    acq_time = instance.AcquisitionTime
    print('New Acquisition location: ' + str(acq_time))

    folder.restore()

def test_set_item_instance():

    folder = Folder(path = test_folder) 
    folder.scan()
    instance = folder.instances(0)

    tags = ['SliceLocation', 'AcquisitionTime', (0x0012, 0x0010)]
    values = instance[tags]
    print('Original Slice location: ' + str(values[0]))
    print('Original Acquisition Time: ' + str(values[1]))
    print('Original Clinical Trial Sponsor: ' + str(values[2]))

    instance[tags] = [0.0, '00:00:00', 'University of Sheffield']
    values = instance[tags]
    print('New Slice location: ' + str(values[0]))
    print('New Acquisition Time: ' + str(values[1]))
    print('New Clinical Trial Sponsor: ' + str(values[2]))

    folder.restore()

def test_set_item():

    folder = Folder(path = test_folder) 
    folder.scan()
    series = folder.series(0)
    instance = series.instances(1)

    tags = ['SliceLocation', 'AcquisitionTime', (0x0012, 0x0010)]
    values = series[tags]
    print('SERIES')
    print('Original Slice location: ' + str(values[0]))
    print('Original Acquisition Time: ' + str(values[1]))
    print('Original Clinical Trial Sponsor: ' + str(values[2]))
    values = instance[tags]
    print('2nd IMAGE')
    print('Original Slice location: ' + str(values[0]))
    print('Original Acquisition Time: ' + str(values[1]))
    print('Original Clinical Trial Sponsor: ' + str(values[2]))

    series[tags] = [0.0, '00:00:00', 'University of Sheffield']
    values = series[tags]
    print('SERIES')
    print('New Slice location: ' + str(values[0]))
    print('New Acquisition Time: ' + str(values[1]))
    print('New Clinical Trial Sponsor: ' + str(values[2]))
    values = instance[tags]
    print('2nd IMAGE')
    print('New Slice location: ' + str(values[0]))
    print('New Acquisition Time: ' + str(values[1]))
    print('New Clinical Trial Sponsor: ' + str(values[2]))    

    folder.restore()

def test_copy_remove_instance():

    folder = Folder(path = test_folder) 
    folder.scan()
    folder.print()
    
    print('****************************************')
    print('MOVE AND COPY FROM ONE SERIES TO ANOTHER')
    print('****************************************')
    
    parent = folder.patients(0).studies(0).series(7)
    instance = parent.instances(0)
    new_parent = folder.patients(0).studies(0).series(6)

    # copy the instance to the new parent and remove the original
    copy = instance.copy_to(new_parent)
    instance.remove()
    folder.print()

    # move the copy back to the original parent
    # this should restore the situation
    copy.move_to(parent)
    folder.print()

    print('')
    print('***************************************')
    print('MOVE AND COPY FROM ONE STUDY TO ANOTHER')
    print('***************************************')

    # take an instance from the first study
    # and create a new study for the copies
    parent = folder.patients(0).studies(0)
    instance = parent.series(7).instances(0)
    new_parent = folder.patients(0).new_child()

    # copy the instance to the new parent and remove the original
    # this is the same as instance.move_to(new_parent) but slower
    copy = instance.copy_to(new_parent)
    instance.remove()
    folder.print()

    # move the copy back to the original parent
    # this restores the original situation
    copy.move_to(instance.parent)
    folder.print()

    folder.restore()

def test_create():

    folder = Folder(path = test_folder) 
    folder.scan()  
    patient = folder.new_child()
    study = patient.new_child()
    series = study.new_child()
    instance = series.new_child()
    print(instance.UID)

def test_copy_remove():

    folder = Folder(path = test_folder) 
    folder.scan()
    folder.print() 
    # create a new study and copy the first series to it.
    study = folder.patients(0).new_child()
    folder.series(0).copy_to(study)
    folder.print()
    folder.restore()

def test_merge():



    # first create two new patients
    folder = Folder(path = test_folder) 
    folder.scan()

    print('')
    print('***************************************')
    print('            CREATE 2 NEW PATIENTS      ')
    print('***************************************')

    patient1 = folder.new_child()
    patient2 = folder.new_child()
    folder.patients(0).series(0).copy_to(patient1)
    folder.patients(0).series(1).copy_to(patient2)
    folder.print()

    print('')
    print('***************************************')
    print('            MERGE 2 PATIENTS           ')
    print('***************************************')

    # then merge the two new patients into a third
    patients_to_merge = [patient1, patient2]
    patient3 = folder.new_child()
    for patient in patients_to_merge:
        for study in patient.studies():
            study.copy_to(patient3)
    folder.print()


    print('')
    print('***************************************')
    print('            MERGE ALL STUDIES          ')
    print('***************************************')

    # now merge all studies of the new patient 
    # into a new study of the same patient.
    studies_to_merge = patient3.studies()
    new_study = patient3.new_child()
    for study in studies_to_merge:
        for series in study.series():
            series.copy_to(new_study)
    folder.print()

    print('')
    print('***************************************')
    print('            MERGE ALL SERIES          ')
    print('***************************************')

    # now merge all series of the new patient into
    # a new series in a new study of the same patient
    series_to_merge = patient3.series()
    new_study = patient3.new_child()
    new_series = new_study.new_child()
    for series in series_to_merge:
        for instance in series.instances():
            instance.copy_to(new_series)
    folder.print()

    folder.restore()


def test_save_restore():

    print('***************************************')
    print('            ORIGINAL FOLDER            ')
    print('***************************************')

    folder = Folder(path = test_folder) 
    folder.scan()
    folder.print()

    # copy the first series into a new patient

    print('***************************************')
    print('        NEW PATIENT CREATED            ')
    print('***************************************')

    patient = folder.new_child()
    folder.series(0).copy_to(patient)
    folder.print()

    # restore and print again

    print('***************************************')
    print('       ORIGINAL FOLDER RESTORED        ')
    print('***************************************')

    folder.restore()
    folder.print()

    # copy the first series into a new patient

    print('***************************************')
    print('       NEW PATIENT CREATED AGAIN       ')
    print('***************************************')

    patient = folder.new_child()
    folder.series(0).copy_to(patient)
    folder.print()

    # save and print again

    print('***************************************')
    print('          NEW PATIENT SAVED            ')
    print('***************************************')

    folder.save()
    folder.print()

def test_read_write_dataset():

    folder = Folder(path = test_folder) 
    folder.scan()

    instance = folder.instances(0)
    ds = instance.read() # work from memory
    print('Original rows and columns')
    print(ds.Rows)
    print(ds.Columns)

    ds.Rows = ds.Rows * 2
    ds.Columns = ds.Columns * 2
    print('Modified rows and columns')
    print(ds.Rows)
    print(ds.Columns)

    matrix = ['Rows','Columns']
    d = instance[matrix]
    instance[matrix] = [int(d[0]*2), int(d[1]*2)]
    print('Modified rows and columns again')
    print(ds.Rows)
    print(ds.Columns)
    
    ds = instance.read() # read original values from disk again
    print('Original rows and columns')
    print(ds.Rows)
    print(ds.Columns)
    ds.Rows = ds.Rows * 2
    ds.Columns = ds.Columns * 2
    instance.write()
    instance.clear()

    instance.read() # read original values from disk again
    print('Modified rows and columns')
    print(instance.Rows)
    print(instance.Columns)
    instance.clear()

    print('Modified rows and columns') #  read from disk
    print(instance.Rows)
    print(instance.Columns)

    folder.restore()    

def test_export():

    folder = Folder(path = test_folder) 
    folder.scan()
    instance = folder.instances(0)
    instance.export(test_export_folder)
    series = folder.series(0)
    series.export(test_export_folder)

def test_SOPClass():

    folder = Folder(path = test_folder) 
    folder.scan()
    for instance in folder.instances():
        print(instance.__class__.__name__)


def test_checking():

    folder = Folder(path = test_folder) 
    folder.open()
    folder.check()
    folder.uncheck()
    series = folder.series(1)
    print('The second series in the folder')
    print(series.label())
    series.check()
    folder.series(0).check()
    folder.save()
    print('The second checked series in the folder')
    series = folder.series(checked=True)
    print(series[1].label())
    print('The second of the checked series')
    series = folder.series(1, checked=True)
    print(series.label())
    instance = series.instances(0)
    instance.uncheck()
    series = folder.series(checked=True)
    print('Nr of checked series after unchecking an instance')
    print(len(series))


def test_all():

    test_read_and_open()
    test_retrieve()
    test_find()
    test_read_item_instance()
    test_read_item()
    test_set_attr_instance()
    test_set_item_instance()
    test_set_item()
    test_create()
    test_copy_remove_instance()
    test_copy_remove()
    test_merge()
    test_save_restore()
    test_read_write_dataset()
    test_export()
    test_SOPClass() 
    test_checking()
    


# For development purposes
# test_tmp()
# test_df()

# test_all()

# test_read_and_open()
# test_retrieve()
# test_find()
# test_read_item_instance()
# test_read_item()
# test_set_attr_instance()
# test_set_item_instance()
# test_set_item()
# test_copy_remove_instance()
# test_create()
# test_copy_remove()
# test_merge()
# test_save_restore()
# test_read_write_dataset()
# test_export()
# test_SOPClass()
# test_checking()
# test_read_series_array()
# test_sort_series()
# test_set_slice_array_invert()
# test_set_array_invert()
# test_set_array_mip()


"""
This folder contains 3rd party packages that are used in this Weasel project.
You can find a list of these below with their respective version. 
Each one has a folder allocated, which contains their own README.md file with their licensing details (all open-source)

- dcm4che Version 5.23.1

"""


import numpy as np

from .MRimage import MRImage


class EnhancedMRImage(MRImage):
    """Specific methods for the SOPClass EnhancedMR Image Storage"""

    def array(self):

        on_disk = self.on_disk()
        if on_disk: self.read()
        pixelArray = self.ds.pixel_array.astype(np.float32)
        frames = self.ds.PerFrameFunctionalGroupsSequence
        for index, frame in enumerate(frames):
            slice = np.squeeze(pixelArray[index, ...])
            if [0x2005, 0x100E] in self.ds: # 'Philips Rescale Slope'
                slope = self.ds[(0x2005, 0x100E)].value
                intercept = ds[(0x2005, 0x100D)].value
                slice = (slice - intercept) / slope
            else:
                transform = frame.PixelValueTransformationSequence[0]
                slope = float(getattr(transform, 'RescaleSlope', 1)) 
                intercept = float(getattr(transform, 'RescaleIntercept', 0)) 
                slice = slice * slope + intercept
            pixelArray[index, ...] = np.transpose(slice)
        if on_disk: self.clear()
        return pixelArray

    def set_array(self, pixelArray, value_range=None):

        on_disk = self.on_disk()
        if on_disk: self.read()

        pixelArray = self.clip(pixelArray, value_range=value_range)
        pixelArray, slope, intercept = self.scale_to_range(pixelArray, self.ds.BitsAllocated)
        pixelArray = np.transpose(pixelArray, (0, 2, 1))

        maximum = np.amax(pixelArray)
        minimum = np.amin(pixelArray)
        shape = np.shape(pixelArray)

        self.ds.NumberOfFrames = np.shape(pixelArray)[0]
        del self.ds.PerFrameFunctionalGroupsSequence[self.ds.NumberOfFrames:]

        del self.ds[0x2005, 0x100E]  # Delete 'Philips Rescale Slope'
        del self.ds[0x2005, 0x100D]
        self.ds.PixelRepresentation = 0
        self.ds.SmallestImagePixelValue = int(maximum)
        self.ds.LargestImagePixelValue = int(minimum)
        self.ds.RescaleSlope = 1 / slope
        self.ds.RescaleIntercept = - intercept / slope
        self.ds.WindowCenter = (maximum + minimum) / 2
        self.ds.WindowWidth = maximum - minimum
        self.ds.Rows = shape[0]
        self.ds.Columns = shape[1]
        self.ds.PixelData = pixelArray.tobytes()

        if on_disk: 
            self.write()
            self.clear()

    def signal_type(self):
        """Determine if an image is Water, Fat, In-Phase, Out-phase image or None"""

        on_disk = self.on_disk()
        if on_disk: self.read()

        flagWater = False
        flagFat = False
        flagInPhase = False
        flagOutPhase = False
        type = self.ds.MRImageFrameTypeSequence[0]
        if hasattr(type, 'FrameType'):
            type = set(type.FrameType)
        elif hasattr(type, 'ComplexImageComponent'):
            type = set(type.ComplexImageComponent)
        else:
            return flagWater, flagFat, flagInPhase, flagOutPhase
        if set(['W', 'WATER']).intersection(type):
            flagWater = True
        elif set(['F', 'FAT']).intersection(type):
            flagFat = True
        elif set(['IP', 'IN_PHASE']).intersection(type):
            flagInPhase = True
        elif set(['OP', 'OUT_PHASE']).intersection(type):
            flagOutPhase = True

        if on_disk: self.clear()
        return flagWater, flagFat, flagInPhase, flagOutPhase

    def image_type(self):
        """Determine if a dataset is Magnitude, Phase, Real or Imaginary"""

        on_disk = self.on_disk()
        if on_disk: self.read()

        flagMagnitude = []
        flagPhase = []
        flagReal = []
        flagImaginary = []
        for index, singleSlice in enumerate(self.ds.PerFrameFunctionalGroupsSequence):
            sequence = singleSlice.MRImageFrameTypeSequence[0]
            if hasattr(sequence, 'FrameType'):
                type = set(sequence.FrameType)
                if set(['M', 'MAGNITUDE']).intersection(type):
                    flagMagnitude.append(index)
                    continue
                elif set(['P', 'PHASE']).intersection(type):
                    flagPhase.append(index)
                    continue
                elif set(['R', 'REAL']).intersection(type):
                    flagReal.append(index)
                    continue
                elif set(['I', 'IMAGINARY']).intersection(type):
                    flagImaginary.append(index)
                    continue
            if hasattr(sequence, 'ComplexImageComponent'):
                type = set(sequence.ComplexImageComponent)
                if set(['M', 'MAGNITUDE']).intersection(type):
                    flagMagnitude.append(index)
                    continue
                elif set(['P', 'PHASE']).intersection(type):
                    flagPhase.append(index)
                    continue
                elif set(['R', 'REAL']).intersection(type):
                    flagReal.append(index)
                    continue
                elif set(['I', 'IMAGINARY']).intersection(type):
                    flagImaginary.append(index)
                    continue
        if on_disk: self.clear()
        return flagMagnitude, flagPhase, flagReal, flagImaginary

    @property
    def affine_matrix(self):

        on_disk = self.on_disk()
        if on_disk: self.read()
        affineList = list()
        for frame in self.ds.PerFrameFunctionalGroupsSequence:
            affine = affine(
                image_orientation = frame.PlaneOrientationSequence[0].ImageOrientationPatient, 
                image_position = frame.PlanePositionSequence[0].ImagePositionPatient, 
                pixel_spacing = frame.PixelMeasuresSequence[0].PixelSpacing, 
                slice_spacing = frame.PixelMeasuresSequence[0].SpacingBetweenSlices)
            affineList.append(affine)
        if on_disk: self.clear()
        return np.squeeze(np.array(affineList))

    def window(self):
        """Centre and width of the pixel data after applying rescale slope and intercept.
        
        In this case retrieve the centre and width values of the first frame
        """
        on_disk = self.on_disk()
        if on_disk: self.read()
        centre = self.ds.PerFrameFunctionalGroupsSequence[0].FrameVOILUTSequence[0].WindowCenter 
        width = self.ds.PerFrameFunctionalGroupsSequence[0].FrameVOILUTSequence[0].WindowWidth
        if centre is None or width is None:
            array = self.array()
        if centre is None: 
            centre = np.median(array)
        if width is None: 
            p = np.percentile(array, [25, 75])
            width = p[1] - p[0]
        if on_disk: self.clear()
        return centre, width


"""Defines the `Series` class for reading and writing series of DICOM instances.

`Series` provides an interface for reading and sorting images from a series
of instances, extracting subseries of certain type such as magnitude images
in MRI, or mapping one series volume into the coordinate space of another, etc 

    # Example: Get a 3D numpy array from the first series in a folder
    # sorted by slice location and acquisition time

    from dbdicom import Folder

    series = Folder('C:\\Users\\MyName\\MyData\\DICOMtestData').open().series(0)
    array, _ = series.array(['SliceLocation', 'AcquisitionTime'])
"""

__all__ = ['Series']

import numpy as np
import nibabel as nib

from .record import Record


class Series(Record):
    """Programming interface for reading and writing DICOM data series"""

    def __init__(self, folder, UID=[], **attributes):
        """Initialise the series with an instance of Folder.

        Args:
            folder: instance of Folder()
            UID: optional 3-element list with UID's of the patient and study the series is part of.
                The final element is the UID of the series to be created.
                
                If UID is not provided or has less than 3 elements, the missing elements
                are automatically created.
        """
        super().__init__(folder, UID, generation=3, **attributes)

    def label(self, row=None):
        """
        A human-readable label to describe the series.
        """
        if row is None:
            data = self.data()
            if data.empty: return "New Series"
            file = data.index[0]
            descr = data.at[file, 'SeriesDescription']
            nr = data.at[file, 'SeriesNumber']
        else:
            descr = row.SeriesDescription
            nr = row.SeriesNumber
            
        label = '[' + str(nr).zfill(3) + '] ' 
        label += str(descr)
        return label
    
    def map_onto(self, target):
        """Map non-zero pixels onto another series"""

        in_memory = self.in_memory()
        source_images = self.children()
        mapped_series = self.parent.new_child()
        if in_memory: mapped_series.read()
        target_images = target.children() # create record.images() to return children of type image
        for i, target_image in enumerate(target_images):
            target_image.read()
            self.status.progress(i, len(target_images))
            pixel_array = np.zeros((target_image.Rows, target_image.Columns), dtype=np.bool) 
            for j, source_image in enumerate(source_images):
                message = (
                    'Mapping image ' + str(j) + 
                    ' of ' + self.SeriesDescription + 
                    ' to image ' + str(i) + 
                    ' of ' + target.SeriesDescription )
                self.status.message(message)
                if not in_memory: source_image.read()
                array = source_image.map_onto(target_image).array().astype(np.bool)
                np.logical_or(pixel_array, array, out=pixel_array)
            if pixel_array.any():
                mapped_image = target_image.copy_to(mapped_series)
                mapped_image.set_array(pixel_array.astype(np.float32))
                mapped_image.SeriesDescription = self.SeriesDescription
                if not in_memory: mapped_image.write()
        self.status.hide()
        return mapped_series

    def export_as_nifti(self, directory=None, filename=None):
        """Export series as a single Nifty file"""

        if directory is None: 
            directory = self.directory(message='Please select a folder for the nifty data')
        if filename is None:
            filename = self.SeriesDescription
        dicomHeader = nib.nifti1.Nifti1DicomExtension(2, self.instances(0).read())
        pixelArray = np.flipud(np.rot90(np.transpose(self.array())))
        niftiObj = nib.Nifti1Instance(pixelArray, self.instances(0).affine)
        niftiObj.header.extensions.append(dicomHeader)
        nib.save(niftiObj, directory + '/' + filename + '.nii.gz')

    def export_as_csv(self, directory=None, filename=None, columnHeaders=None):
        """Export all images as csv files"""

        if directory is None: 
            directory = self.directory(message='Please select a folder for the csv data')
        if filename is None:
            filename = self.SeriesDescription
        for i, instance in enumerate(self.instances()):
            instance.export_as_csv(self, 
                directory = directory, 
                filename = filename + '(' + str(i) + ')', 
                columnHeaders = columnHeaders)

    def magnitude(self):
        "Creates a sibling series with the magnitude images"

        return self._extractImageType('MAGNITUDE')

    def phase(self):
        "Creates a sibling series with the phase images"

        return self._extractImageType('PHASE')

    def real(self):
        "Creates a sibling series with the real images"

        return self._extractImageType('REAL')

    def imaginary(self):
        "Creates a sibling series with the imaginary images"

        return self._extractImageType('IMAGINARY')

    def _extractImageType(self, type):
        """Extract subseries with images of given imageType"""

        series = self.parent.new_child()
        for instance in self.instances():
            if instance.image_type() == type:
                instance.copy_to(series)
        return series

    def _amax(self, axis=None):
        """Calculate the maximum of the image array along a given dimension.
        
        This function is included as a placeholder reminder 
        to build up functionality at series level that emulates 
        numpy behaviour.

        Args:
            axis: DICOM KeyWord string to specify the dimension
            along which the maximum is taken.

        Returns:
            a new sibling series holding the result.

        Example:
        ```ruby
        # Create a maximum intensity projection along the slice dimension:
        mip = series.amax(axis='SliceLocation')
        ```
        """
        array, data = self.array(axis)
        array = np.amax(array, axis=0)
        data = np.squeeze(data[0,...])
        series = self.new_sibling()
        series.set_array(array, data)
        return series



__all__ = ['QImage']

import os
import struct
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import nibabel as nib
from matplotlib import cm
from nibabel.affines import apply_affine
from PyQt5 import QtGui

from .instance import Instance

class Image(Instance):
    """Specific methods for the SOPClass MR Image Storage"""

    def _initialize(self, ref_ds=None):
        """Initialize the attributes relevant for the Images"""

        super()._initialize(ref_ds)
        
        self.ImageType.insert(0, "DERIVED")

    def array(self):
        """Read the pixel array from an image"""

        on_disk = self.on_disk()
        if on_disk: self.read()
        if self.ds is None: return
        array = self.ds.pixel_array.astype(np.float32)
        slope = float(getattr(self.ds, 'RescaleSlope', 1)) 
        intercept = float(getattr(self.ds, 'RescaleIntercept', 0)) 
        #array = array * slope + intercept
        array *= slope
        array += intercept
        array = np.transpose(array)
        if on_disk: self.clear()
        return array

    def set_array(self, array, value_range=None):

        on_disk = self.on_disk()
        if on_disk: self.read()
        
        if array.ndim >= 3: # remove spurious dimensions of 1
            array = np.squeeze(array) 
        array = self.clip(array, value_range=value_range)
        array, slope, intercept = self.scale_to_range(array, self.ds.BitsAllocated)
        array = np.transpose(array)

        maximum = np.amax(array)
        minimum = np.amin(array)
        shape = np.shape(array)

        self.ds.PixelRepresentation = 0
        self.ds.SmallestImagePixelValue = int(maximum)
        self.ds.LargestImagePixelValue = int(minimum)
        self.ds.RescaleSlope = 1 / slope
        self.ds.RescaleIntercept = - intercept / slope
#        self.ds.WindowCenter = (maximum + minimum) / 2
#        self.ds.WindowWidth = maximum - minimum
        self.ds.Rows = shape[0]
        self.ds.Columns = shape[1]
        self.ds.PixelData = array.tobytes()
        if on_disk: 
            self.write()
            self.clear()

#    def write_array(self, pixelArray, value_range=None): # obsolete - remove
#        """Write the pixel array to disk"""
#        self.set_array(pixelArray, value_range=value_range)
#        self.write()

    def clip(self, array, value_range = None):

        array[np.isnan(array)] = 0
        if value_range is None:
            finite = array[np.isfinite(array)]
            value_range = [np.amin(finite), np.amax(finite)]
        return np.clip(array, value_range[0], value_range[1])

    def scale_to_range(self, array, bits_allocated):
            
    #    target = np.power(2, bits_allocated) - 1
        target = 2.0**bits_allocated - 1
        maximum = np.amax(array)
        minimum = np.amin(array)
        if maximum == minimum:
            slope = 1
        else:
            slope = target / (maximum - minimum)
        intercept = -slope * minimum
        # array = slope * (array - minimum)
        array = slope * array + intercept

        if bits_allocated == 8:
            return array.astype(np.uint8), slope, intercept
        if bits_allocated == 16:
            return array.astype(np.uint16), slope, intercept
        if bits_allocated == 32:
            return array.astype(np.uint32), slope, intercept
        if bits_allocated == 64:
            return array.astype(np.uint64), slope, intercept

    def zeros(self):

        array = np.zeros((self.Rows, self.Columns))
        new = self.copy()
    #    new.write_array(array)
        new.set_array(array)
        new.write()
        return new

    def map_onto(self, target):
        """Map non-zero image pixels onto a target image.
        
        Overwrite pixel values in the target"""

        # Create a coordinate array of non-zero pixels
        coords = np.transpose(np.where(self.array() != 0)) 
        coords = [[coord[0], coord[1], 0] for coord in coords] 
        coords = np.array(coords)

        # Determine coordinate transformation matrix
        affineSource = self.affine_matrix()
        affineTarget = target.affine_matrix()
        sourceToTarget = np.linalg.inv(affineTarget).dot(affineSource)

        # Apply coordinate transformation
        coords = apply_affine(sourceToTarget, coords)
        coords = np.round(coords, 3).astype(int)
        x = tuple([coord[0] for coord in coords if coord[2] == 0])
        y = tuple([coord[1] for coord in coords if coord[2] == 0])

        # Set values in the target image
        # Note - replace by actual values rather than 1 & 0.
        result = target.zeros()
        in_memory = self.in_memory()
        if in_memory: result.read()
        pixelArray = result.array()
        pixelArray[(x, y)] = 1.0
        result.set_array(pixelArray)
        if not in_memory: 
            result.write()
            result.clear()

        return result

    def affine_matrix(self):
        """Affine transformation matrix for a DICOM image"""

        on_disk = self.on_disk()
        if on_disk: self.read()

        image_orientation = self.ds.ImageOrientationPatient
        image_position = self.ds.ImagePositionPatient
        pixel_spacing = self.ds.PixelSpacing
        slice_spacing = self.ds.SliceThickness            

        row_spacing = pixel_spacing[0]
        column_spacing = pixel_spacing[1]

        row_cosine = np.array(image_orientation[:3])
        column_cosine = np.array(image_orientation[3:])
        slice_cosine = np.cross(row_cosine, column_cosine)

        affine = np.identity(4, dtype=np.float32)
        affine[:3, 0] = row_cosine * column_spacing
        affine[:3, 1] = column_cosine * row_spacing
        affine[:3, 2] = slice_cosine * slice_spacing
        affine[:3, 3] = image_position

        if on_disk: self.clear()

        return affine

    def get_colormap(self):
        """Returns the colormap if there is any."""

        on_disk = self.on_disk()
        if on_disk: self.read()
        ds = self.ds

        lut = None
        if hasattr(ds, 'ContentLabel'):
            if ds.PhotometricInterpretation == 'PALETTE COLOR':
                colormap = ds.ContentLabel
            elif 'MONOCHROME' in ds.PhotometricInterpretation:
                colormap = 'gray'
        elif len(ds.dir("PaletteColor"))>=3 and ds.PhotometricInterpretation == 'PALETTE COLOR':
            colormap = 'custom'
            lut = self.get_lut()
        else:
            colormap = 'gray' # default

        if on_disk: self.clear()
        return colormap, lut  

    def get_lut(self):
        
        on_disk = self.on_disk()
        if on_disk: self.read()
        ds = self.ds

        redColour = list(ds.RedPaletteColorLookupTableData)
        greenColour = list(ds.GreenPaletteColorLookupTableData)
        blueColour = list(ds.BluePaletteColorLookupTableData)
        redLut = list(struct.unpack('<' + ('H' * ds.RedPaletteColorLookupTableDescriptor[0]), bytearray(redColour)))
        greenLut = list(struct.unpack('<' + ('H' * ds.GreenPaletteColorLookupTableDescriptor[0]), bytearray(greenColour)))
        blueLut = list(struct.unpack('<' + ('H' * ds.BluePaletteColorLookupTableDescriptor[0]), bytearray(blueColour)))
        colours = np.transpose([redLut, greenLut, blueLut])
        normaliseFactor = int(np.power(2, ds.RedPaletteColorLookupTableDescriptor[2]))
        # Fast ColourTable loading
        colourTable = np.around(colours/normaliseFactor, decimals = 2)
        indexes = np.unique(colourTable, axis=0, return_index=True)[1]
        lut = [colourTable[index].tolist() for index in sorted(indexes)]
        # Full / Complete Colourmap - takes 20 seconds to load each image
        # lut = (colours/normaliseFactor).tolist()   
        if on_disk: self.clear()
        return lut      

    def set_colormap(self, colormap=None, levels=None):
        """Set the colour table of the image."""

        on_disk = self.on_disk()
        if on_disk: self.read()
        ds = self.ds

        #and (colormap != 'gray') removed from If statement below, so as to save gray colour tables
        if (colormap == 'gray'):
            ds.PhotometricInterpretation = 'MONOCHROME2'
            ds.ContentLabel = ''
            if hasattr(ds, 'RedPaletteColorLookupTableData'):
                del (ds.RGBLUTTransferFunction, ds.RedPaletteColorLookupTableData,
                    ds.GreenPaletteColorLookupTableData, ds.BluePaletteColorLookupTableData,
                    ds.RedPaletteColorLookupTableDescriptor, ds.GreenPaletteColorLookupTableDescriptor,
                    ds.BluePaletteColorLookupTableDescriptor)
        if ((colormap is not None)  and (colormap != 'custom') and (colormap != 'gray') 
            and (colormap != 'default') and isinstance(colormap, str)):
            ds.PhotometricInterpretation = 'PALETTE COLOR'
            ds.RGBLUTTransferFunction = 'TABLE'
            ds.ContentLabel = colormap
            stringType = 'US' # ('SS' if minValue < 0 else 'US')
            ds.PixelRepresentation = 0 # (1 if minValue < 0 else 0)
            pixelArray = ds.pixel_array
            minValue = int(np.amin(pixelArray))
            maxValue = int(np.amax(pixelArray))
            numberOfValues = int(maxValue - minValue)
            arrayForRGB = np.arange(0, numberOfValues)
            colorsList = cm.ScalarMappable(cmap=colormap).to_rgba(np.array(arrayForRGB), bytes=False)
            totalBytes = ds.BitsAllocated
            ds.add_new('0x00281101', stringType, [numberOfValues, minValue, totalBytes])
            ds.add_new('0x00281102', stringType, [numberOfValues, minValue, totalBytes])
            ds.add_new('0x00281103', stringType, [numberOfValues, minValue, totalBytes])
            ds.RedPaletteColorLookupTableData = bytes(np.array([int((np.power(
                2, totalBytes) - 1) * value) for value in colorsList[:, 0].flatten()]).astype('uint'+str(totalBytes)))
            ds.GreenPaletteColorLookupTableData = bytes(np.array([int((np.power(
                2, totalBytes) - 1) * value) for value in colorsList[:, 1].flatten()]).astype('uint'+str(totalBytes)))
            ds.BluePaletteColorLookupTableData = bytes(np.array([int((np.power(
                2, totalBytes) - 1) * value) for value in colorsList[:, 2].flatten()]).astype('uint'+str(totalBytes)))
        if levels is not None:
            ds.WindowCenter = levels[0]
            ds.WindowWidth = levels[1]
        if on_disk: self.clear()

    def export_as_nifti(self, directory=None, filename=None):
        """Export 2D pixel Array in nifty format"""

        on_disk = self.on_disk()
        if on_disk: self.read()
        ds = self.ds
        if directory is None: 
            directory = self.directory(message='Please select a folder for the nifty data')
        if filename is None:
            filename = self.SeriesDescription
        dicomHeader = nib.nifti1.Nifti1DicomExtension(2, ds)
        niftiObj = nib.Nifti1Instance(np.flipud(np.rot90(np.transpose(self.array()))), affine=self.affine)
        # The transpose is necessary in this case to be in line with the rest of WEASEL.
        niftiObj.header.extensions.append(dicomHeader)
        nib.save(niftiObj, directory + '/' + filename + '.nii.gz')
        if on_disk: self.clear()

    def export_as_csv(self, directory=None, filename=None, columnHeaders=None):
        """Export 2D pixel Array in csv format"""

        if directory is None: 
            directory = self.directory(message='Please select a folder for the csv data')
        if filename is None:
            filename = self.SeriesDescription
        filename = os.path.join(directory, filename + '.csv')
        table = self.array()
        if columnHeaders is None:
            columnHeaders = []
            counter = 0
            for _ in table:
                counter += 1
                columnHeaders.append("Column" + str(counter))
        df = pd.DataFrame(np.transpose(table), columns=columnHeaders)
        df.to_csv(filename, index=False) 

    def export_as_png(self, fileName):
        """Export image in png format."""

        colourTable, _ = self.get_colormap()
        pixelArray = np.transpose(self.array())
        centre, width = self.window()
        minValue = centre - width/2
        maxValue = centre + width/2
        cmap = plt.get_cmap(colourTable)
        plt.imshow(pixelArray, cmap=cmap)
        plt.clim(int(minValue), int(maxValue))
        cBar = plt.colorbar()
        cBar.minorticks_on()
        plt.savefig(fname=fileName + '_' + self.label() + '.png')
        plt.close()

    def window(self):
        """Centre and width of the pixel data after applying rescale slope and intercept"""

        on_disk = self.on_disk()
        if on_disk: self.read()
        if 'WindowCenter' in self.ds: centre = self.ds.WindowCenter
        if 'WindowWidth' in self.ds: width = self.ds.WindowWidth
        if centre is None or width is None:
            array = self.array()
        if centre is None: 
            centre = np.median(array)
        if width is None: 
            p = np.percentile(array, [25, 75])
            width = p[1] - p[0]
        if on_disk: self.clear()
        return centre, width

    def QImage(self):

        array = self.array()
        return QImage(array, width=self.WindowWidth, center=self.WindowCenter)


def QImage(array, width=None, center=None):

    if (width is None) or (center is None):
        max = np.amax(array)
        min = np.amin(array)
    if width is None: width = max-min
    if center is None: center = (max-min)/2

    imgData, alpha = _makeARGB(
        data = array, 
        levels = [center-width/2, center+width/2],
    )
    return _makeQImage(imgData, alpha)


# HELPER FUNCTIONS ADAPTED FROM pyQtGraph


def _makeARGB(data, lut=None, levels=None, scale=None, useRGBA=False): 
    """ 
    Convert an array of values into an ARGB array suitable for building QImages
    
    Returns the ARGB array (unsigned byte) and a boolean indicating whether
    there is alpha channel data. This is a two stage process:
    
        1) Rescale the data based on the values in the *levels* argument (min, max).
        2) Determine the final output by passing the rescaled values through a
           lookup table.
   
    Both stages are optional.
    
    ============== ==================================================================================
    **Arguments:**
    data           numpy array of int/float types. If 
    levels         List [min, max]; optionally rescale data before converting through the
                   lookup table. The data is rescaled such that min->0 and max->*scale*::
                   
                      rescaled = (clip(data, min, max) - min) * (*scale* / (max - min))
                   
                   It is also possible to use a 2D (N,2) array of values for levels. In this case,
                   it is assumed that each pair of min,max values in the levels array should be 
                   applied to a different subset of the input data (for example, the input data may 
                   already have RGB values and the levels are used to independently scale each 
                   channel). The use of this feature requires that levels.shape[0] == data.shape[-1].
    scale          The maximum value to which data will be rescaled before being passed through the 
                   lookup table (or returned if there is no lookup table). By default this will
                   be set to the length of the lookup table, or 255 if no lookup table is provided.
    lut            Optional lookup table (array with dtype=ubyte).
                   Values in data will be converted to color by indexing directly from lut.
                   The output data shape will be input.shape + lut.shape[1:].
                   Lookup tables can be built using ColorMap or GradientWidget.
    useRGBA        If True, the data is returned in RGBA order (useful for building OpenGL textures). 
                   The default is False, which returns in ARGB order for use with QImage 
                   (Note that 'ARGB' is a term used by the Qt documentation; the *actual* order 
                   is BGRA).
    ============== ==================================================================================
    """

    if data.ndim not in (2, 3):
        raise TypeError("data must be 2D or 3D")
    if data.ndim == 3 and data.shape[2] > 4:
        raise TypeError("data.shape[2] must be <= 4")
    
    if lut is not None and not isinstance(lut, np.ndarray):
        lut = np.array(lut)
    
    if levels is None:
        # automatically decide levels based on data dtype
        if data.dtype.kind == 'u':
            levels = np.array([0, 2**(data.itemsize*8)-1])
        elif data.dtype.kind == 'i':
            s = 2**(data.itemsize*8 - 1)
            levels = np.array([-s, s-1])
        elif data.dtype.kind == 'b':
            levels = np.array([0,1])
        else:
            raise Exception('levels argument is required for float input types')
    if not isinstance(levels, np.ndarray):
        levels = np.array(levels)
    if levels.ndim == 1:
        if levels.shape[0] != 2:
            raise Exception('levels argument must have length 2')
    elif levels.ndim == 2:
        if lut is not None and lut.ndim > 1:
            raise Exception('Cannot make ARGB data when both levels and lut have ndim > 2')
        if levels.shape != (data.shape[-1], 2):
            raise Exception('levels must have shape (data.shape[-1], 2)')
    else:
        raise Exception("levels argument must be 1D or 2D (got shape=%s)." % repr(levels.shape))

    # Decide on maximum scaled value
    if scale is None:
        if lut is not None:
            scale = lut.shape[0] - 1
        else:
            scale = 255.

    # Decide on the dtype we want after scaling
    if lut is None:
        dtype = np.ubyte
    else:
        dtype = np.min_scalar_type(lut.shape[0]-1)
            
    # Apply levels if given
    if levels is not None:
        if isinstance(levels, np.ndarray) and levels.ndim == 2:
            # we are going to rescale each channel independently
            if levels.shape[0] != data.shape[-1]:
                raise Exception("When rescaling multi-channel data, there must be the same number of levels as channels (data.shape[-1] == levels.shape[0])")
            newData = np.empty(data.shape, dtype=int)
            for i in range(data.shape[-1]):
                minVal, maxVal = levels[i]
                if minVal == maxVal:
                    maxVal += 1e-16
                newData[...,i] = _rescaleData(data[...,i], scale/(maxVal-minVal), minVal, dtype=dtype)
            data = newData
        else:
            # Apply level scaling unless it would have no effect on the data
            minVal, maxVal = levels
            if minVal != 0 or maxVal != scale:
                if minVal == maxVal:
                    maxVal += 1e-16
                data = _rescaleData(data, scale/(maxVal-minVal), minVal, dtype=dtype)
            
    # apply LUT if given
    if lut is not None:
        data = _applyLookupTable(data, lut)
    else:
        if data.dtype is not np.ubyte:
            data = np.clip(data, 0, 255).astype(np.ubyte)

    # this will be the final image array
    imgData = np.empty(data.shape[:2]+(4,), dtype=np.ubyte)

    # decide channel order
    if useRGBA:
        order = [0,1,2,3] # array comes out RGBA
    else:
        order = [2,1,0,3] # for some reason, the colors line up as BGR in the final image.
        
    # copy data into image array
    if data.ndim == 2:
        # This is tempting:
        #   imgData[..., :3] = data[..., np.newaxis]
        # ..but it turns out this is faster:
        for i in range(3):
            imgData[..., i] = data
    elif data.shape[2] == 1:
        for i in range(3):
            imgData[..., i] = data[..., 0]
    else:
        for i in range(0, data.shape[2]):
            imgData[..., i] = data[..., order[i]] 
    
    # add opaque alpha channel if needed
    if data.ndim == 2 or data.shape[2] == 3:
        alpha = False
        imgData[..., 3] = 255
    else:
        alpha = True

    return imgData, alpha


def _makeQImage(imgData, alpha=None, copy=True, transpose=True):
    """
    Turn an ARGB array into QImage. 'd
    By default, the data is copied; changes to the array will not
    be reflected in the image. The image will be given aata' attribute
    pointing to the array which shares its data to prevent python
    freeing that memory while the image is in use.
    
    ============== ===================================================================
    **Arguments:**
    imgData        Array of data to convert. Must have shape (width, height, 3 or 4) 
                   and dtype=ubyte. The order of values in the 3rd axis must be 
                   (b, g, r, a).
    alpha          If True, the QImage returned will have format ARGB32. If False,
                   the format will be RGB32. By default, _alpha_ is True if
                   array.shape[2] == 4.
    copy           If True, the data is copied before converting to QImage.
                   If False, the new QImage points directly to the data in the array.
                   Note that the array must be contiguous for this to work
                   (see numpy.ascontiguousarray).
    transpose      If True (the default), the array x/y axes are transposed before 
                   creating the image. Note that Qt expects the axes to be in 
                   (height, width) order whereas pyqtgraph usually prefers the 
                   opposite.
    ============== ===================================================================    
    """
    ## create QImage from buffer
    
    ## If we didn't explicitly specify alpha, check the array shape.
    if alpha is None:
        alpha = (imgData.shape[2] == 4)
        
    copied = False
    if imgData.shape[2] == 3:  ## need to make alpha channel (even if alpha==False; QImage requires 32 bpp)
        if copy is True:
            d2 = np.empty(imgData.shape[:2] + (4,), dtype=imgData.dtype)
            d2[:,:,:3] = imgData
            d2[:,:,3] = 255
            imgData = d2
            copied = True
        else:
            raise Exception('Array has only 3 channels; cannot make QImage without copying.')
    
    if alpha:
        imgFormat = QtGui.QImage.Format_ARGB32
    else:
        imgFormat = QtGui.QImage.Format_RGB32
        
    if transpose:
        imgData = imgData.transpose((1, 0, 2))  ## QImage expects the row/column order to be opposite

    if not imgData.flags['C_CONTIGUOUS']:
        if copy is False:
            extra = ' (try setting transpose=False)' if transpose else ''
            raise Exception('Array is not contiguous; cannot make QImage without copying.'+extra)
        imgData = np.ascontiguousarray(imgData)
        copied = True
        
    if copy is True and copied is False:
        imgData = imgData.copy()       
    try:
        img = QtGui.QImage(imgData.ctypes.data, imgData.shape[1], imgData.shape[0], imgFormat)
    except:
        img = QtGui.QImage(memoryview(imgData), imgData.shape[1], imgData.shape[0], imgFormat)
                
    img.data = imgData
    
    return img
    
def _applyLookupTable(data, lut):
    """
    Uses values in *data* as indexes to select values from *lut*.
    The returned data has shape data.shape + lut.shape[1:]
    """
    if data.dtype.kind not in ('i', 'u'):
        data = data.astype(int)
    
    return np.take(lut, data, axis=0, mode='clip')  


def _rescaleData(data, scale, offset, dtype=None, clip=None):
    """Return data rescaled and optionally cast to a new dtype::
    
        data => (data-offset) * scale
        
    """
    if dtype is None:
        dtype = data.dtype
    else:
        dtype = np.dtype(dtype)
    
    try:
        newData = np.empty((data.size,), dtype=dtype)
        flat = np.ascontiguousarray(data).reshape(data.size)
        newData = (flat - offset)*scale 
        if dtype != dtype: 
            newData = newData.astype(dtype)
        data = newData.reshape(data.shape)
    except:
        
        d2 = data - float(offset)
        d2 *= scale
        
        # Clip before converting dtype to avoid overflow
        if dtype.kind in 'ui':
            lim = np.iinfo(dtype)
            if clip is None:
                # don't let rescale cause integer overflow
                d2 = np.clip(d2, lim.min, lim.max)
            else:
                d2 = np.clip(d2, max(clip[0], lim.min), min(clip[1], lim.max))
        else:
            if clip is not None:
                d2 = np.clip(d2, *clip)
        data = d2.astype(dtype)
    return data



from .classes.database import Database
from .classes.patient import Patient
from .classes.study import Study
from .classes.series import Series
from .classes.instance import Instance
from .classes.MRimage import MRImage
from .classes.enhancedMRimage import EnhancedMRImage
from .classes.secondary_capture_image import SecondaryCaptureImage

def series(folder, uid, **attributes):
    return Series(folder, UID=uid, **attributes)

def instance(folder, uid, SOPClassUID=None, **attributes):
    """Creates an instance of a dicm object from a row in the dataframe"""

    if SOPClassUID is None:
        return Instance(folder, UID=uid, **attributes)
    if SOPClassUID == '1.2.840.10008.5.1.4.1.1.4': 
        return MRImage(folder, UID=uid, **attributes)
    if SOPClassUID == '1.2.840.10008.5.1.4.1.1.4.1': 
        return EnhancedMRImage(folder, UID=uid, **attributes)
    if SOPClassUID == '1.2.840.10008.5.1.4.1.1.7': 
        return SecondaryCaptureImage(folder, UID=uid, **attributes)  

def object(folder, row=None, generation=4, **attributes):
    """Creates an instance of a dicm object from a row in the dataframe"""

    if generation == 0: 
        return Database(folder, UID=[], **attributes)

    key = folder._columns[0:generation]
    UID = row[key].values.tolist()

    if generation == 1: 
        return Patient(folder, UID=UID, **attributes)
    if generation == 2: 
        return Study(folder, UID=UID, **attributes)
    if generation == 3: 
        return Series(folder, UID=UID, **attributes)
    if generation == 4: 
        if row.SOPClassUID == '1.2.840.10008.5.1.4.1.1.4': 
            return MRImage(folder, UID=UID, **attributes)
        if row.SOPClassUID == '1.2.840.10008.5.1.4.1.1.4.1': 
            return EnhancedMRImage(folder, UID=UID, **attributes)
        if row.SOPClassUID == '1.2.840.10008.5.1.4.1.1.7': 
            return SecondaryCaptureImage(folder, UID=UID, **attributes)
        else: 
            return Instance(folder, UID=UID, **attributes)

def new_child(obj, **attributes):
    """Creates a new child object for a DICOM object"""

    if obj.generation == 0:
        return Patient(obj.folder, UID=obj.UID, **attributes)
    if obj.generation == 1:
        return Study(obj.folder, UID=obj.UID, **attributes)
    if obj.generation == 2:
        return Series(obj.folder, UID=obj.UID, **attributes)
    if obj.generation == 3: 
        if obj._SOPClassUID == '1.2.840.10008.5.1.4.1.1.4': 
            return MRImage(obj.folder, UID=obj.UID, **attributes)
        if obj._SOPClassUID == '1.2.840.10008.5.1.4.1.1.4.1': 
            return EnhancedMRImage(obj.folder, UID=obj.UID, **attributes)
        if obj._SOPClassUID == '1.2.840.10008.5.1.4.1.1.7': 
            return SecondaryCaptureImage(obj.folder, UID=obj.UID, **attributes)
        else:
            return Instance(obj.folder, UID=obj.UID, **attributes)
    if obj.generation == 4: 
        return None

def parent(obj):
    "Returns the parent object"

    if obj.generation == 0: 
        return None
    if obj.generation == 1:
        return Database(obj.folder, UID=obj.UID[:-1])
    if obj.generation == 2:
        return Patient(obj.folder, UID=obj.UID[:-1])
    if obj.generation == 3:
        return Study(obj.folder, UID=obj.UID[:-1])
    if obj.generation == 4:
        return Series(obj.folder, UID=obj.UID[:-1])


import os
import sys
import zipfile
from datetime import datetime
import math
import pathlib
import subprocess
import pydicom
from pydicom.dataset import Dataset
from pydicom.sequence import Sequence
import pandas as pd
import numpy as np

def dataframe(path, files, tags, status=None, message='Reading DICOM folder..'):
    """Reads a list of tags in a list of files.

    Arguments
    ---------
    files : str or list
        A filepath or a list of filepaths
    tags : str or list 
        A DICOM tag or a list of DICOM tags
    status : StatusBar

    Creates
    -------
    dataframe : pandas.DataFrame
        A Pandas dataframe with one row per file
        The index is the file path 
        Each column corresponds to a Tag in the list of Tags
        The returned dataframe is sorted by the given tags.
    """
    if not isinstance(files, list):
        files = [files]
    if not isinstance(tags, list):
        tags = [tags]
    array = []
    dicom_files = []
    if status is not None: status.message(message)
    for i, file in enumerate(files):
        ds = pydicom.dcmread(file, force=True)
        if isinstance(ds, pydicom.dataset.FileDataset):
            if 'TransferSyntaxUID' in ds.file_meta:
                row = _read_tags(ds, tags)
                array.append(row)
                relpath = os.path.relpath(file, path)
                dicom_files.append(relpath) 
        if status is not None: status.progress(i+1, len(files), message)
    if status is not None: status.hide()
    return pd.DataFrame(array, index = dicom_files, columns = tags)

def _initialize(ds, UID=None, ref=None):

    # Date and Time of Creation
    dt = datetime.now()
    timeStr = dt.strftime('%H%M%S')  # long format with micro seconds

    ds.ContentDate = dt.strftime('%Y%m%d')
    ds.ContentTime = timeStr
    ds.AcquisitionDate = dt.strftime('%Y%m%d')
    ds.AcquisitionTime = timeStr
    ds.SeriesDate = dt.strftime('%Y%m%d')
    ds.SeriesTime = timeStr
    ds.InstanceCreationDate = dt.strftime('%Y%m%d')
    ds.InstanceCreationTime = timeStr

    if UID is not None:

        # overwrite UIDs
        ds.PatientID = UID[0]
        ds.StudyInstanceUID = UID[1]
        ds.SeriesInstanceUID = UID[2]
        ds.SOPInstanceUID = UID[3]

    if ref is not None: 

        # Series, Instance and Class for Reference
        refd_instance = Dataset()
        refd_instance.ReferencedSOPClassUID = ref.SOPClassUID
        refd_instance.ReferencedSOPInstanceUID = ref.SOPInstanceUID
        refd_instance_sequence = Sequence()
        refd_instance_sequence.append(refd_instance)

        refd_series = Dataset()
        refd_series.ReferencedInstanceSequence = refd_instance_sequence
        refd_series.SeriesInstanceUID = ds.SeriesInstanceUID
        refd_series_sequence = Sequence()
        refd_series_sequence.append(refd_series)

        ds.ReferencedSeriesSequence = refd_series_sequence

    return ds

def _unzip_files(path, status):
    """
    Unzip any zipped files in a directory.
    
    Checking for zipped files is slow so this only searches the top folder.

    Returns : a list with unzipped files
    """
    files = [entry.path for entry in os.scandir(path) if entry.is_file()]
    zipfiles = []
    for i, file in enumerate(files):
        status.progress(i, len(files), message='Searching for zipped folders..')
        if zipfile.is_zipfile(file):
            zipfiles.append(file)
    if zipfiles == []:
        return
    for i, file in enumerate(zipfiles): # unzip any zip files and delete the original zip
        status.progress(i, len(zipfiles), 'Unzipping file ' + file)
        with zipfile.ZipFile(file, 'r') as zip_folder:
            path = ''.join(file.split('.')[:-1]) # remove file extension
            if not os.path.isdir(path): 
                os.mkdir(path)
            zip_folder.extractall(path)
        os.remove(file)

def _read_tags(ds, tags):
    """Helper function return a list of values"""

    # https://pydicom.github.io/pydicom/stable/guides/element_value_types.html
    if not isinstance(tags, list): 
        if tags not in ds:
            return None
        else:
        #    return ds[tags].value
            return _convert_attribute_type(ds[tags].value)
            
    row = []  
    for tag in tags:
        if tag not in ds:
            value = None
        else:
        #    value = ds[tag].value
            value = _convert_attribute_type(ds[tag].value)
        row.append(value)
    return row

def _convert_attribute_type(value):
    """Convert pyidcom datatypes to the python datatypes used to set the parameter.
    
    While this removes some functionality, this aligns with the principle 
    of `dbdicom` to remove DICOM-native langauge from the API.
    """

    if value.__class__.__name__ == 'PersonName':
        return str(value)
    if value.__class__.__name__ == 'Sequence':
        return [ds for ds in value]
    if value.__class__.__name__ == 'TM': # This can probably do with some formatting
        return str(value) 
    if value.__class__.__name__ == 'UID': 
        return str(value) 
    if value.__class__.__name__ == 'IS': 
        return int(value)
    if value.__class__.__name__ == 'DT': 
        return str(value)
    if value.__class__.__name__ == 'DA': 
        return str(value)
    if value.__class__.__name__ == 'DSfloat': 
        return float(value)
    if value.__class__.__name__ == 'DSdecimal': 
        return int(value)
    else:
        return value

def _set_tags(ds, tags, values):
    """Sets DICOM tags in the dataset in memory"""

    if not isinstance(tags, list): 
        tags = [tags]
        values = [values]
    for i, tag in enumerate(tags):
        if tag in ds:
            ds[tag].value = values[i]
        else:
            if not isinstance(tag, pydicom.tag.BaseTag):
                tag = pydicom.tag.Tag(tag)
            if not tag.is_private: # Add a new data element
                VR = pydicom.datadict.dictionary_VR(tag)
                ds.add_new(tag, VR, values[i])
            else:
                pass # for now
    return ds

def _filter(objects, **kwargs):
    """
    Filters a list of DICOM classes by DICOM tags and values.
    
    Example
    -------
    instances = _filter(instances, PatientName="Harry")
    """

    filtered = []
    for obj in objects:
        select = True
        for tag, value in kwargs.items():
            if getattr(obj, tag) != value:
                select = False
                break
        if select: 
            filtered.append(obj)
    return filtered



def split_multiframe(filepath, description):
    """Splits a multi-frame instance into single frames"""

    multiframeDir = os.path.dirname(filepath)
    fileBase = "SingleFrame_"
    fileBaseFlag = fileBase + "000000_" + description.replace('.', '_')
    command = [program('emf2sf'), "--inst-no", "'%s'", "--not-chseries", "--out-dir", multiframeDir, "--out-file", fileBaseFlag, filepath]
    try:
        fail = subprocess.call(command, stdout=subprocess.PIPE)
    except Exception as e:
        fail = 1
        print('Error in dcm4che: Could not split the detected Multi-frame DICOM file.\n'\
                'The DICOM file ' + filepath + ' was not deleted.')

    # Return a list of newly created files
    multiframe_files_list = []
    if fail == 0:
        for new_file in os.listdir(multiframeDir):
            if new_file.startswith(fileBase):
                new_file_path = os.path.join(multiframeDir, new_file)
                multiframe_files_list.append(new_file_path)     
                # Slice Locations need to be copied from a private field 
                ds = pydicom.dcmread(new_file_path, force=True)
                ds.SliceLocation = ds[0x2001,0x100a].value
                ds.save_as(new_file_path)
    return multiframe_files_list

def program(script):
    """Helper function: Find the program for a script"""

    if os.name =='nt': 
        script += '.bat'
    program = script
    # If running Weasel as executable
    if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
        directory = pathlib.Path(sys._MEIPASS)
    # If running Weasel as normal Python script
    else:
        directory = pathlib.Path().absolute()
    for dirpath, _, filenames in os.walk(directory):
        for filename in filenames:
            if filename.endswith(script):
                sys.path.append(dirpath)
                program = os.path.join(dirpath, filename)
    return program

def scan_tree(directory):
    """Helper function: yield DirEntry objects for the directory."""

    for entry in os.scandir(directory):
        if entry.is_dir(follow_symlinks=False):
            yield from scan_tree(entry.path)
        else:
            yield entry

def _stack_arrays(arrays, align_left=False):
    """Stack a list of arrays of different shapes but same number of dimensions.
    
    This generalises numpy.stack to arrays of different sizes.
    The stack has the size of the largest array.
    If an array is smaller it is zero-padded and centred on the middle.
    """

    # Get the dimensions of the stack
    # For each dimension, look for the largest values across all arrays
    ndim = len(arrays[0].shape)
    dim = [0] * ndim
    for array in arrays:
        for i, d in enumerate(dim):
            dim[i] = max((d, array.shape[i])) # changing the variable we are iterating over!!
    #    for i in range(ndim):
    #        dim[i] = max((dim[i], array.shape[i]))

    # Create the stack
    # Add one dimension corresponding to the size of the stack
    n = len(arrays)
    #stack = np.full([n] + dim, 0, dtype=arrays[0].dtype)
    stack = np.full([n] + dim, None, dtype=arrays[0].dtype)

    for k, array in enumerate(arrays):
        index = [k]
        for i, d in enumerate(dim):
            if align_left:
                i0 = 0
            else: # align center and zero-pad missing values
                i0 = math.floor((d-array.shape[i])/2)
            i1 = i0 + array.shape[i]
            index.append(slice(i0,i1))
        stack[tuple(index)] = array

    return stack



import os
from copy import deepcopy
import pydicom
import numpy as np
import pandas as pd
from .. import utilities
import dbdicom as db


class Record():

    def __init__(self, folder, UID=[], generation=0, **attributes):

        objUID = [] + UID
#        for i in range(generation-len(UID)):
        while generation > len(objUID):
            newUID = pydicom.uid.generate_uid()
            objUID.append(newUID)    

        self.__dict__['UID'] = objUID
        self.__dict__['folder'] = folder
        self.__dict__['status'] = folder.status
        self.__dict__['dialog'] = folder.dialog
        self.__dict__['dicm'] = folder.dicm
        self.__dict__['ds'] = None
        # placeholder DICOM attributes
        # these will populate the dataset and dataframe when data are created
        self.__dict__['attributes'] = attributes

    @property
    def generation(self):
        return len(self.UID)

    @property
    def key(self):
        """The keywords describing the UID of the record"""

        key = ['PatientID', 'StudyInstanceUID', 'SeriesInstanceUID', 'SOPInstanceUID']
        return key[0:self.generation]

    def new_uid(self):
        
        return pydicom.uid.generate_uid()

    def data(self):
        """Dataframe with current data - excluding those that were removed
        """

        # Note: this returns a copy - could be a view instead using .loc?

        if self.folder.path is None:
            return self.folder.dataframe
        current = self.folder.dataframe.removed == False
        data = self.folder.dataframe[current]
        if self.UID == []: 
            return data       
        rows = data[self.key[-1]] == self.UID[-1]
        return data[rows]

    def dataset(self, sortby=None, status=True): 
        """Sort instances by a list of attributes.
        
        Args:
            sortby: 
                List of DICOM keywords by which the series is sorted
        Returns:
            An ndarray holding the instances sorted by sortby.
        """
        if sortby is None:
            df = self.data()
            return self._dataset_from_df(df)
        else:
            if set(sortby) <= set(self.folder.dataframe):
                df = self.folder.dataframe.loc[self.data().index, sortby]
            else:
                df = utilities.dataframe(self.folder.path, self.files, sortby, self.status)
            df.sort_values(sortby, inplace=True) 
            return self._sorted_dataset_from_df(df, sortby, status=status)

    def _sorted_dataset_from_df(self, df, sortby, status=True): 

        data = []
        vals = df[sortby[0]].unique()
        for i, c in enumerate(vals):
            if status: self.status.progress(i, len(vals), message='Sorting..')
            dfc = df[df[sortby[0]] == c]
            if len(sortby) == 1:
                datac = self._dataset_from_df(dfc)
            else:
                datac = self._sorted_dataset_from_df(dfc, sortby[1:], status=False)
            data.append(datac)
        return utilities._stack_arrays(data, align_left=True)

    def _dataset_from_df(self, df): 
        """Return datasets as numpy array of object type"""

        data = np.empty(df.shape[0], dtype=object)
        cnt = 0
        for file, _ in df.iterrows(): # just enumerate over df.index
            #self.status.progress(cnt, df.shape[0])
            data[cnt] = self.folder.instance(file)
            cnt += 1
        #self.status.hide()
        return data

    def array(self, sortby=None, pixels_first=False): 
        """Pixel values of the object as an ndarray
        
        Args:
            sortby: 
                Optional list of DICOM keywords by which the volume is sorted
            pixels_first: 
                If True, the (x,y) dimensions are the first dimensions of the array.
                If False, (x,y) are the last dimensions - this is the default.

        Returns:
            An ndarray holding the pixel data.

            An ndarry holding the datasets (instances) of each slice.

        Examples:
            ``` ruby
            # return a 3D array (z,x,y)
            # with the pixel data for each slice
            # in no particular order (z)
            array, _ = series.array()    

            # return a 3D array (x,y,z)   
            # with pixel data in the leading indices                               
            array, _ = series.array(pixels_first = True)    

            # Return a 4D array (x,y,t,k) sorted by acquisition time   
            # The last dimension (k) enumerates all slices with the same acquisition time. 
            # If there is only one image for each acquision time, 
            # the last dimension is a dimension of 1                               
            array, data = series.array('AcquisitionTime', pixels_first=True)                         
            v = array[:,:,10,0]                 # First image at the 10th location
            t = data[10,0].AcquisitionTIme      # acquisition time of the same image

            # Return a 4D array (loc, TI, x, y) 
            sortby = ['SliceLocation','InversionTime']
            array, data = series.array(sortby) 
            v = array[10,6,0,:,:]            # First slice at 11th slice location and 7th inversion time    
            Loc = data[10,6,0][sortby[0]]    # Slice location of the same slice
            TI = data[10,6,0][sortby[1]]     # Inversion time of the same slice
            ```  
        """
        dataset = self.dataset(sortby)
        array = []
        ds = dataset.ravel()
        for i, im in enumerate(ds):
            self.status.progress(i, len(ds), 'Reading pixel data..')
            if im is None:
                array.append(np.zeros((1,1)))
            else:
                array.append(im.array())
        self.status.hide()
        #array = [im.array() for im in dataset.ravel() if im is not None]
        array = utilities._stack_arrays(array) #db.stack(array)
        array = array.reshape(dataset.shape + array.shape[1:])
        if pixels_first:
            array = np.moveaxis(array, -1, 0)
            array = np.moveaxis(array, -1, 0)
        return array, dataset # REPLACE BY DBARRAY

    def npy(self):

        path = os.path.join(self.folder.path, "dbdicom_npy")
        if not os.path.isdir(path): os.mkdir(path)
        file = os.path.join(path, self.UID[-1] + '.npy') 
        return file

    def load_npy(self):

        file = self.npy()
        if not os.path.exists(file):
            return
        with open(file, 'rb') as f:
            array = np.load(f)
        return array

    def save_npy(self, array=None, sortby=None, pixels_first=False):

        if array is None:
            array = self.array(sortby=sortby, pixels_first=pixels_first)
        file = self.npy() 
        with open(file, 'wb') as f:
            np.save(f, array)

    def set_array(self, array, dataset=None, pixels_first=False, inplace=False): 
        """
        Set pixel values of a series from a numpy ndarray.

        Since the pixel data do not hold any information about the 
        image such as geometry, or other metainformation,
        a dataset must be provided as well with the same 
        shape as the array except for the slice dimensions. 

        If a dataset is not provided, header info is 
        derived from existing instances in order.

        Args:
            array: 
                numpy ndarray with pixel data.

            dataset: 
                numpy ndarray

                Instances holding the header information. 
                This *must* have the same shape as array, minus the slice dimensions.

            pixels_first: 
                bool

                Specifies whether the pixel dimensions are the first or last dimensions of the series.
                If not provided it is assumed the slice dimensions are the last dimensions
                of the array.

            inplace: 
                bool

                If True (default) the current pixel values in the series 
                are overwritten. If set to False, the new array is added to the series.
        
        Examples:
            ```ruby
            # Invert all images in a series:
            array, _ = series.array()
            series.set_array(-array)

            # Create a maximum intensity projection of the series.
            # Header information for the result is taken from the first image.
            # Results are saved in a new sibling series.
            array, data = series.array()
            array = np.amax(array, axis=0)
            data = np.squeeze(data[0,...])
            series.new_sibling().set_array(array, data)

            # Create a 2D maximum intensity projection along the SliceLocation direction.
            # Header information for the result is taken from the first slice location.
            # Current data of the series are overwritten.
            array, data = series.array('SliceLocation')
            array = np.amax(array, axis=0)
            data = np.squeeze(data[0,...])
            series.set_array(array, data)

            # In a series with multiple slice locations and inversion times,
            # replace all images for each slice location with that of the shortest inversion time.
            array, data = series.array(['SliceLocation','InversionTime']) 
            for loc in range(array.shape[0]):               # loop over slice locations
                slice0 = np.squeeze(array[loc,0,0,:,:])     # get the slice with shortest TI 
                TI0 = data[loc,0,0].InversionTime           # get the TI of that slice
                for TI in range(array.shape[1]):            # loop over TIs
                    array[loc,TI,0,:,:] = slice0            # replace each slice with shortest TI
                    data[loc,TI,0].InversionTime = TI0      # replace each TI with shortest TI
            series.set_array(array, data)
            ```
        """
        if pixels_first:    # Move to the end (default)
            array = np.moveaxis(array, 0, -1)
            array = np.moveaxis(array, 0, -1)
        if dataset is None:
            dataset = self.dataset()
        # Return with error message if dataset and array do not match.
        nr_of_slices = np.prod(array.shape[:-2])
        if nr_of_slices != np.prod(dataset.shape):
            message = 'Error in set_array(): array and dataset do not match'
            message += '\n Array has ' + str(nr_of_slices) + ' elements'
            message += '\n dataset has ' + str(np.prod(dataset.shape)) + ' elements'
            message += '\n Check if the keyword pixels_first is set correctly.'
            self.dialog.error(message)
            raise ValueError(message)
        # If self is not a series, create a new series.
        if self.generation != 3:
            series = self.new_series()
        else:
            series = self
        # Reshape, copy instances and save slices.
        array = array.reshape((nr_of_slices, array.shape[-2], array.shape[-1])) # shape (i,x,y)
        dataset = dataset.reshape(nr_of_slices) # shape (i,)

        dataset = db.copy(dataset.tolist(), series, status=self.status)
        for i, instance in enumerate(dataset):
            self.status.progress(i, len(dataset), 'Writing array to file..')
            instance.set_array(array[i,...])
            if inplace: instance.remove() # delete?
           
        #for i, instance in enumerate(dataset):
        #    self.status.progress(i, len(dataset), 'Saving data..')
        #    instance.copy_to(series).set_array(array[i,...])
            # instance.set_array(array[i,...])
        #    if inplace: instance.remove() # delete?

        return series

#    def write_array(self, array, dataset): 
#        """
#        Set and array and write it to disk.
#        """
#        series = self.set_array(array, dataset)
#        series.write()
#        return series

    @property
    def _SOPClassUID(self):
        """The SOP Class UID of the first instance"""

        data = self.data()
        if data.empty: return None
        return self.data().iloc[0].SOPClassUID

    @property
    def files(self):
        """Returns the filepath to the instances in the object."""
 
        relpaths = self.data().index.tolist()
        return [os.path.join(self.folder.path, p) for p in relpaths]

    def in_memory(self): # is_in_memory
        """Check if the object has been read into memory"""

        return self.ds is not None

    def on_disk(self): # is_on_disk

        return self.ds is None

    @property
    def parent(self):
        "Returns the parent object"

        return self.dicm.parent(self)
        
    def children(self, index=None, **kwargs):
        """List of children"""

        if self.generation == 4: return []
        if self.in_memory():
            objects = utilities._filter(self.ds, **kwargs)
            if index is not None:
                if index >= len(objects): 
                    return
                else:
                    return objects[index]
            return objects
        return self.records(generation=self.generation+1, index=index, **kwargs)

    def records(self, generation=0, index=None, **kwargs):
        """A list of all records of a given generation corresponding to the record.

        If generation is lower then that of the object, 
        all offspring of the given generation are returned.

        If the generation is higher than that of the object,
        the correspondong ancestor is return as a 1-element list.

        Optionally the list can be filtered by index, or by providing a 
        list of DICOM KeyWords and values. In that case only objects
        a returned that fulfill all criteria.
        
        Parameters
        ----------
        generation : int
            The generation to be returned (0 to 4)
        index : int
            Index of the single object to be return
        kwargs : (Key, Value)
            Conditions to filter the objects
        """
        objects = []
        if generation == 0:
            obj = self.dicm.object(self.folder, generation=0)
            objects.append(obj)
        else:
            key = self.folder._columns[0:generation]
            data = self.data()
            if data.empty: 
                if index is None:
                    return objects
                else:
                    return
            column = data[key[-1]]
            rec_list = column.unique()
            if index is not None:
                rec_list = [rec_list[index]]
            for rec in rec_list:
                rec_data = data[column == rec]
                row = rec_data.iloc[0]
                obj = self.dicm.object(self.folder, row, generation)
                objects.append(obj)
        objects = utilities._filter(objects, **kwargs)
        if index is not None: return objects[0]
        return objects

    def patients(self, index=None,  **kwargs):
        """A list of patients of the object"""

        if self.generation==4: 
            return self.parent.parent.parent
        if self.generation==3:
            return self.parent.parent
        if self.generation==2:
            self.parent
        if self.generation==1:
            return
        return self.children(index=index, **kwargs)

    def studies(self, index=None, **kwargs):
        """A list of studies of the object"""

        if self.generation==4: 
            return self.parent.parent
        if self.generation==3:
            return self.parent
        if self.generation==2:
            return
        if self.generation==1:
            return self.children(index=index, **kwargs)
        objects = []
        for child in self.children():
            inst = child.studies(**kwargs)
            objects.extend(inst)
        if index is not None:
            if index >= len(objects):
                return
            else:
                return objects[index]
        return objects

    def series(self, index=None, **kwargs):
        """A list of series of the object"""

        if self.generation==4: 
            return self.parent
        if self.generation==3:
            return
        if self.generation==2:
            kids = self.children(index=index, **kwargs)
            return kids
        series = []
        for child in self.children():
            inst = child.series(**kwargs)
            series.extend(inst)
        if index is not None:
            if index >= len(series):
                return
            else:
                return series[index]
        return series

    def instances(self, index=None, **kwargs): # VERY slow - needs optimizing
        """A list of instances of the object"""

        if self.generation==4: 
            return
        if self.generation==3:
            return self.children(index=index, **kwargs)
        instances = []
        for child in self.children():
            inst = child.instances(**kwargs)
            instances.extend(inst)
        if index is not None:
            if index >= len(instances):
                return
            else:
                return instances[index]
        return instances       

    def new_child(self, **attributes):
        """Creates a new child object"""

        obj = self.dicm.new_child(self, **attributes)
        obj.read() # Why??
        return obj

    def new_sibling(self, **attributes):
        """
        Creates a new sibling under the same parent.
        """
        if self.generation == 0:
            return
        else:
            return self.parent.new_child(**attributes)

    def new_pibling(self, **attributes):
        """
        Creates a new sibling of parent.
        """
        if self.generation <= 1:
            return
        else:
            return self.parent.new_sibling(**attributes)

    def new_cousin(self, **attributes):
        """
        Creates a new sibling of parent.
        """
        if self.generation <= 1:
            return
        else:
            return self.new_pibling().new_child(**attributes)

    def new_series(self, **attributes):
        """
        Creates a new series under the same parent
        """ 
        if self.generation <= 1: 
            return self.new_child().new_series(**attributes)
        if self.generation == 2:
            return self.new_child(**attributes)
        if self.generation == 3:
            return self.new_sibling(**attributes)
        if self.generation == 4:
            return self.new_pibling(**attributes) 

    def __getattr__(self, tag):
        """Gets the value of the data element with given tag.
        
        Arguments
        ---------
        tag : str
            DICOM KeyWord String

        Returns
        -------
        Value of the corresponding DICOM data element
        """
        return self[tag]

    def __setattr__(self, tag, value):
        """Sets the value of the data element with given tag."""

        if tag == 'folder':
            self.__dict__['folder'] = value
        elif tag == 'ds':
            self.__dict__['ds'] = value
        else:
            self[tag] = value

    def __getitem__(self, tags):
        """Gets the value of the data elements with specified tags.
        
        Arguments
        ---------
        tags : a string, hexadecimal tuple, or a list of strings and hexadecimal tuples

        Returns
        -------
        A value or a list of values
        """
        instance = self.instances(0)
        if instance is not None:
            return instance[tags]

    def __setitem__(self, tags, values):
        """Sets the value of the data element with given tag."""

        # FASTER but needs testing
        # This also means __setitem__ can be removed from instance class
        #
        # db.set_value(self.instances(), dict(zip(tags, values)))
        
        # LAZY - SLOW
        instances = self.instances()
        self.status.message('Writing DICOM tags..')
        for i, instance in enumerate(instances):
            instance[tags] = values
            self.status.progress(i, len(instances))
        self.status.hide()

    def remove(self):
        """Deletes the object. """ 

        files = self.files
        if files == []: 
            return
        self.folder.dataframe.loc[self.data().index,'removed'] = True


    def move_to(self, ancestor):
        """move object to a new parent.
        
        ancestor:any DICOM Class
            If the object is not a parent, the missing 
            intermediate generations are automatically created.
        """
        copy = self.copy_to(ancestor)
        self.remove()
        return copy

#    def move(self, child, ancestor):
#        """Move a child object to a new parent"""

#        if self.in_memory():
#            if child in self.ds:
#                self.ds.remove(child)
#        child = child.move_to(ancestor)
    
    def copy(self):
        """Returns a copy in the same parent"""

        copy = self.copy_to(self.parent)
        if self.in_memory(): copy.read()
        return copy

    def _copy_to_OBSOLETE(self, ancestor, message=None): # functional but slow
        """copy object to a new ancestor.
        
        ancestor: Root, Patient or Study
        If the object is not a study, the missing 
        intermediate generations are automatically created.
        """
        if self.generation == 0: return
#        if ancestor.generation == 0: return
        copy = self.__class__(self.folder, UID=ancestor.UID)
        if ancestor.in_memory():
            copy.read()
            ancestor.ds.append(copy)
        children = self.children()
        if message is None:
            message = "Copying " + self.__class__.__name__ + ' ' + self.label()
        self.status.message(message)
        for i, child in enumerate(children):
            child.copy_to(copy)
            self.status.progress(i, len(children))
        self.status.hide()
        return copy

    def _initialize(self, ref=None):

        if self.generation == 4:
            self.ds = utilities.initialize(self.ds, UID=self.UID, ref=ref)
        else:
            for i, obj in enumerate(self.ds):
                if ref is not None:
                    obj._initialize(ref.ds[i])
                else:
                    obj._initialize()

    def merge_with(self, obj, message=None): 

        if self.generation == 0: return
        if message is None:
            message = "Merging " + self.__class__.__name__ + ' ' + self.label()
        # replace by db.merge()
        return self._merge_with(obj, obj.parent, message=message)

    def copy_to(self, ancestor, message=None):

        if self.generation == 0: return
        if message is None:
            message = "Copying " + self.__class__.__name__ + ' ' + self.label()
        copy = self.__class__(self.folder, UID=ancestor.UID)
        # replace by db.merge()
        return self._merge_with(copy, ancestor, message=message)

    def _merge_with(self, obj, ancestor, message=None): # obsolete - replace by db.merge()

        if self.in_memory(): # Create the copy in memory
            obj.__dict__['ds'] = deepcopy(self.ds)
            obj._initialize(self.ds)
            if ancestor.in_memory():
                if ancestor.generation == obj.generation-1:
                    ancestor.ds.append(obj)
            return obj

        # Extend dataframe & create new files
        dfsource = self.data()
        sourcefiles = [os.path.join(self.folder.path, p) for p in dfsource.index.tolist()]
        df = dfsource.copy(deep=True)
        df['files'] = [self.folder.new_file() for _ in range(df.shape[0])]
        df.set_index('files', inplace=True)
        for key in self.folder._columns[self.generation:3]:
            for id in df[key].unique():
                uid = self.folder.new_uid()
                rows = df[key] == id
                df.loc[rows.index, key] = uid
            #    for file in rows.index:
            #        df.at[file, key] = uid
        df.SOPInstanceUID = self.folder.new_uid(df.shape[0])
        df.removed = False
        df.created = True
        copyfiles = df.index.tolist()

        for i, file in enumerate(copyfiles):
            self.status.progress(i, len(copyfiles), message=message)
            df.loc[file, self.folder._columns[0:self.generation]] = obj.UID 
            ds = pydicom.dcmread(sourcefiles[i])
            ds = utilities._initialize(ds, UID=df.loc[file, self.folder._columns[:4]].values.tolist())
            if obj.attributes is not None:
                for key, value in obj.attributes.items():
                    utilities._set_tags(ds, key, value)
                    if key in self.folder._columns[4:]:
                        df.at[file, key] = value  
            ds.save_as(os.path.join(self.folder.path, file))

        self.folder.__dict__['dataframe'] = pd.concat([self.folder.dataframe, df])
        self.status.hide()

        if ancestor.in_memory():
            if ancestor.generation == obj.generation-1:
                obj.read() # unnecessary read - can be integrated in loop.
                ancestor.ds.append(obj)

        return obj

    def export(self, path):
        """Export instances to an external folder.

        The instance itself will not be removed from the DICOM folder.
        Instead a copy of the file will be copied to the external folder.
        
        Arguments
        ---------
        path : str
            path to an external folder. If not provided,
            a window will prompt the user to select one.
        """
        instances = self.instances()
        self.status.message('Exporting..')
        for i, instance in enumerate(instances):
            instance.export(path)
            self.status.progress(i,len(instances))
        self.status.hide()

    def save_OBSOLETE(self):
        """Save all instances of the record."""

        # Slow - edit df directly
        self.status.message("Saving all current instances..")
        instances = self.instances() 
        for i, instance in enumerate(instances):
            instance.save()
            self.status.progress(i, len(instances))
        self.status.hide()

        self.status.message("Deleting all removed instances..")
        if self.__class__.__name__ == 'Folder':
            data = self.folder.dataframe
        else:
            rows = self.folder.dataframe[self.key[-1]] == self.UID[-1]
            data = self.folder.dataframe[rows] 
        removed = data.removed[data.removed]
        files = [os.path.join(self.folder.path, p) for p in removed.index.tolist()]
        for i, file in enumerate(files): 
            os.remove(file)
            self.status.progress(i, len(files))
        self.folder.dataframe.drop(removed.index, inplace=True)
        self.status.hide()

    def save(self, message = "Saving changes.."):
        """Save all instances of the record."""

        self.status.message(message)
        self.write()
        if self.generation == 0:
            data = self.folder.dataframe
        else:
            rows = self.folder.dataframe[self.key[-1]] == self.UID[-1]
            data = self.folder.dataframe[rows] 

        created = data.created[data.created]   
        removed = data.removed[data.removed]

        files = [os.path.join(self.folder.path, p) for p in removed.index.tolist()]
        for i, file in enumerate(files): 
            self.status.progress(i, len(files), message='Deleting removed files..')
            if os.path.exists(file): os.remove(file)
        #self.status.message('Clearing rapid access storage..')
        #npyfile = self.npy()
        #if os.path.exists(npyfile): os.remove(npyfile)
        self.status.message('Done saving..')
        self.folder.dataframe.loc[created.index, 'created'] = False
        self.folder.dataframe.drop(removed.index, inplace=True)

    def restore(self, message = "Restoring saved state.."):
        """
        Restore all instances.
        """
        self.status.message(message)

        in_memory = self.in_memory() 
        self.clear()

        if self.generation == 0:
            data = self.folder.dataframe
        else:
            rows = self.folder.dataframe[self.key[-1]] == self.UID[-1]
            data = self.folder.dataframe[rows] 
        created = data.created[data.created]   
        removed = data.removed[data.removed]

        files = [os.path.join(self.folder.path, p) for p in created.index.tolist()]
        for i, file in enumerate(files): 
            self.status.progress(i, len(files), message='Deleting new files..')
            if os.path.exists(file): os.remove(file)
        self.status.hide()
        self.folder.dataframe.loc[removed.index, 'removed'] = False
        self.folder.dataframe.drop(created.index, inplace=True)

        if in_memory: self.read()
        return self
        
    def restore_OBSOLETE(self, message = 'Restoring..'):
        """
        Restore all instances.
        """
        in_memory = self.in_memory() 
        self.clear()

        instances = self.instances()
        self.status.message(message)
        for i, instance in enumerate(instances):
            instance.restore()
            self.status.progress(i,len(instances))
        self.status.hide()

        if in_memory: self.read()
        return self

    def read_dataframe(self, tags):

        return utilities.dataframe(self.folder.path, self.files, tags, self.status)

    def read(self, message = 'Reading..'):

        self.status.message(message)
        self.__dict__['ds'] = self.children()
        for i, child in enumerate(self.ds):
            self.status.progress(i, len(self.ds))
            child.read()
        self.status.hide()

    def write(self):

        if self.ds is None: 
            return
        for i, child in enumerate(self.ds):
            self.status.progress(i, len(self.ds), message='Writing data..')
            child.write()
        self.status.hide()

    def clear(self):

        if self.ds is None: 
            return
        for child in self.ds:
            child.clear()
        self.ds = None



from .record import Record

class Patient(Record):

    def __init__(self, folder, UID=[], **attributes):
        super().__init__(folder, UID, generation=1, **attributes)


    def label(self, row=None):

        if row is None:
            data = self.data()
            if data.empty: return "New Patient"
            file = data.index[0]
            name = data.at[file, 'PatientName']
            id = data.at[file, 'PatientID']
        else:
            name = row.PatientName
            id = row.PatientID
            
        label = str(name)
        label += ' [' + str(id) + ']'
        return label




__all__ = ['set_value', 'merge', 'write', 'read', 'find_series', 'move', 'copy']

import os
from copy import deepcopy
import pandas as pd
import numpy as np
import pydicom
from . import utilities

def write(ds, file, dialog=None):

    try:
        ds.save_as(file) 
    except:
        message = "Failed to write to " + file
        message += "\n The file may be open in another application, or is being synchronised by a cloud service."
        message += "\n Please close the file or pause the synchronisation and try again."
        if dialog is not None:
            dialog.information(message) 
        else:
            print(message)  

def read(file, dialog=None):

    try:
        return pydicom.dcmread(file)
    except:
        message = "Failed to read " + file
        message += "\n The file may be open in another application, or is being synchronised by a cloud service."
        message += "\n Please close the file or pause the synchronisation and try again."
        if dialog is not None:
            dialog.information(message) 
        else:
            print(message)           
    

# This needs to be generalized to objects of any type
def set_value(instances, status=None, **kwargs):
    """Set value of a DICOM tag"""

    if not isinstance(instances, list):
        instances = [instances]
    dialog = instances[0].dialog
    folder = instances[0].folder
    
    if status is not None: 
        status.message('Finding instances in register..')

    # Find the instances in the dataframe. 
    # Separate those that are changing for the first time 
    # from those that have been changed before
    uids = [i.UID[-1] for i in instances]
    df = folder.dataframe
    df = df.loc[df.SOPInstanceUID.isin(uids) & (df.removed == False)]
    df_first_change = df.loc[df.created == False]
    df_prevs_change = df.loc[df.created == True]

    if status is not None: 
        status.message('Extending register with new files..')

    # Create a copy of those that are changing for the first time.
    # Allocate new filenames and append to the dataframe.
    df_created = df_first_change.copy(deep=True)
    df_created.removed = False
    df_created.created = True
    df_created['file'] = [folder.new_file() for _ in range(df_created.shape[0])]
    df_created.set_index('file', inplace=True)
    folder.__dict__['dataframe'] = pd.concat([folder.dataframe, df_created])

    # Mark the original files for removal.
    folder.dataframe.loc[df_first_change.index, 'removed'] = True

    if status is not None: 
        status.message('Writing values to files..')

    # Read all the datasets, update the attributes, write the results to disk.
    # Either in the same file (if the dataset has been changed before)
    # Or in a new file (if this is the first change)
    n = df_prevs_change.shape[0]+df_first_change.shape[0]
    cnt = 1
    for i, filename in enumerate(df_prevs_change.index.values):
        if status is not None: status.progress(cnt, n)
        file = os.path.join(folder.path, filename)
        ds = read(file, dialog)
        for tag, value in kwargs.items():
            ds = utilities._set_tags(ds, tag, value)
            if tag in folder._columns:
                folder.dataframe.loc[filename, tag] = value
        write(ds, file, dialog)
        cnt+=1
    for i, filename in enumerate(df_first_change.index.values):
        if status is not None: status.progress(cnt, n)
        file = os.path.join(folder.path, filename)
        newfilename = df_created.index.values[i]
        newfile = os.path.join(folder.path, newfilename)
        ds = read(file, dialog)
        for tag, value in kwargs.items():
            ds = utilities._set_tags(ds, tag, value)
            if tag in folder._columns:
                folder.dataframe.loc[newfilename, tag] = value
        write(ds, newfile, dialog)
        cnt+=1

    if status is not None: 
        status.message('Finished changing values..')

def copy(instances, series=None, status=None):

    if isinstance(instances, list):
        pass
    elif isinstance(instances, np.ndarray):
        instances = instances.ravel()
    else:   # scalar value
        instances = [instances]
    instances = [i.__class__(i.folder, UID=i.UID) for i in instances if i is not None]
    if instances == []:
        return []
    dialog = instances[0].dialog
    folder = instances[0].folder
    if status is not None: 
        status.message('Moving datasets..')
    if series is None:
        series = instances[0].new_pibling(SeriesDescription='Copy')

    attributes = {}
    attributes['PatientID'] = series.UID[0]
    attributes['StudyInstanceUID'] = series.UID[1]
    attributes['SeriesInstanceUID'] = series.UID[2]
    if series.attributes is not None:
        attributes.update(series.attributes)
        #for key, value in series.attributes.items():
        #    attributes[key] = value

    # Find the instances in the dataframe. 
    uids = [i.UID[-1] for i in instances]
    df = folder.dataframe
    df = df.loc[df.SOPInstanceUID.isin(uids) & (df.removed == False)]
    instance_ind = []
    for filename in df.index.values:
        uid = df.at[filename, 'SOPInstanceUID']
        instance_ind.append(uids.index(uid))

    if status is not None: 
        status.message('Extending register with new files..')

    # Create a copy of the df
    # Allocate new filenames and append to the dataframe.
    df_created = df.copy(deep=True)
    df_created.removed = False
    df_created.created = True
    df_created['file'] = [folder.new_file() for _ in range(df_created.shape[0])]
    df_created.set_index('file', inplace=True)
    folder.__dict__['dataframe'] = pd.concat([folder.dataframe, df_created])

    if status is not None: 
        status.message('Writing values to files..')

    # Read all the datasets, update the attributes, 
    # write the results in the new file
    cnt, n = 1, df.shape[0]
    for i, filename in enumerate(df.index.values):
        if status is not None: 
            status.progress(cnt, n)
        file = os.path.join(folder.path, filename)
        newfilename = df_created.index.values[i]
        newfile = os.path.join(folder.path, newfilename)
        ds = read(file, dialog)
        for tag, value in attributes.items():
            ds = utilities._set_tags(ds, tag, value)
            if tag in folder._columns:
                folder.dataframe.loc[newfilename, tag] = value
        uid = folder.new_uid()
        ds = utilities._set_tags(ds, 'SOPInstanceUID', uid)
        folder.dataframe.loc[newfilename, 'SOPInstanceUID'] = uid
        instances[instance_ind[i]].UID[-1] = uid
        instances[instance_ind[i]].UID[:-1] = series.UID
        write(ds, newfile, dialog)
        cnt+=1

    if status is not None: 
        status.message('Finished changing values..')
    
    return instances

def move(instances, series=None, status=None):

    if status is not None: 
        status.message('Moving datasets..')
    if series is None:
        series = instances[0].new_pibling(SeriesDescription='Copy')
    set_value(instances, 
        status = status,
        PatientID = series.UID[0],
        StudyInstanceUID = series.UID[1],
        SeriesInstanceUID = series.UID[2],
    )
    return instances

def merge(series_list, merged=None, status=None):

    if status is not None: 
        status.message('Merging..')
    if merged is None:
        merged = series_list[0].new_cousin(
            SeriesDescription='Merged series',
        )
    instances = find_instances(series_list)
#    instances = []
#    for i, series in enumerate(series_list):
#        status.progress(i, len(series_list), 'Reading instances..')
#        instances += series.instances() # slow - go via df
    set_value(instances, 
        status = status,
        PatientID = merged.UID[0],
        StudyInstanceUID = merged.UID[1],
        SeriesInstanceUID = merged.UID[2],
    )
    return merged

# Gneralize this to other parents & **kwargs
def find_instances(series, status=None):

    if status is not None:
        status.message('Finding instances..')
    if not isinstance(series, list):
        series = [series]
    UID = [s.UID[-1] for s in series]
    df = series[0].folder.dataframe
    df = df.loc[(df.removed == False) & (df.SeriesInstanceUID.isin(UID))]
    key = series[0].folder._columns[:4]
    all_instances = []
    cnt=0
    for _, row in df.iterrows():
        if status is not None:
            status.progress(cnt, df.shape[0], 'Finding instances..')
        instance = series[0].dicm.instance(
            series[0].folder, 
            row[key].values.tolist(),
            SOPClassUID = row.SOPClassUID)
        all_instances.append(instance)
        cnt+=1
    return all_instances

# Gneralize this to **kwargs but searching df where feasible
def find_series(studies, SeriesDescription=[], status=None):

    if status is not None:
        status.message('Finding series..')
    if not isinstance(studies, list):
        studies = [studies]
    if not isinstance(SeriesDescription, list):
        SeriesDescription = [SeriesDescription]
    UID = [s.UID[-1] for s in studies]
    df = studies[0].folder.dataframe
    select = (df.removed == False) & (df.StudyInstanceUID.isin(UID))
    if SeriesDescription != []:
        desc = SeriesDescription[0]
        series_select = (df.SeriesDescription == desc)
        for desc in SeriesDescription[1:]:
            series_select = series_select | (df.SeriesDescription == desc)
        select = select & series_select
    df = df.loc[select]
    key = studies[0].folder._columns[:3]
    unique_series = df.SeriesInstanceUID.unique()
    all_series = []
    cnt=0
    for uid in unique_series:
        if status is not None:
            status.progress(cnt, len(unique_series), 'Finding series..')
        row = df.loc[df.SeriesInstanceUID == uid].iloc[0]
        UID = row[key].values.tolist()
        series = studies[0].dicm.series(studies[0].folder, UID)
        all_series.append(series)
        cnt+=1
    return all_series
    


import struct
import numpy as np
from .image import Image

class MRImage(Image):
    """Specific methods for the SOPClass MR Image Storage"""

    def array(self):
        """Read the pixel array from an MR image"""

        on_disk = self.on_disk()
        if on_disk: self.read()
        if [0x2005, 0x100E] in self.ds: # 'Philips Rescale Slope'
            array = self.ds.pixel_array.astype(np.float32)
            slope = self.ds[(0x2005, 0x100E)].value
            intercept = self.ds[(0x2005, 0x100D)].value
            array -= intercept
            array /= slope
            array = np.transpose(array)
        else:
            array = super().array()
        if on_disk: self.clear()
        return array
        
    def set_array(self, pixelArray, value_range=None):

        on_disk = self.on_disk()
        if on_disk: self.read()
        if self.ds is None:
            # TODO: Handle this by creating new dataset from scratch
            raise RuntimeError('Cannot set array: no dataset defined on disk or in memory')
        if (0x2005, 0x100E) in self.ds: del self.ds[0x2005, 0x100E]  # Delete 'Philips Rescale Slope'
        if (0x2005, 0x100D) in self.ds: del self.ds[0x2005, 0x100D]
        super().set_array(pixelArray, value_range=value_range)
        #self.write()
        if on_disk: 
            self.write()
            self.clear()

    def image_type(self):
        """Determine if an image is Magnitude, Phase, Real or Imaginary image or None"""

        on_disk = self.on_disk()
        if on_disk: self.read()
        if (0x0043, 0x102f) in self.ds:
            private_ge = self.ds[0x0043, 0x102f]
            try: value = struct.unpack('h', private_ge.value)[0]
            except: value = private_ge.value
            if value == 0: return 'MAGNITUDE'
            if value == 1: return 'PHASE'
            if value == 2: return 'REAL'
            if value == 3: return 'IMAGINARY'
        if 'ImageType' in self.ds:
            type = set(self.ds.ImageType)
            if set(['M', 'MAGNITUDE']).intersection(type):
                return 'MAGNITUDE'
            if set(['P', 'PHASE']).intersection(type):
                return 'PHASE'
            if set(['R', 'REAL']).intersection(type):
                return 'REAL'
            if set(['I', 'IMAGINARY']).intersection(type):
                return 'IMAGINARY'
        if 'ComplexImageComponent' in self.ds:
            return self.ds.ComplexImageComponent
        if on_disk: self.clear()

    def signal_type(self):
        """Determine if an image is Water, Fat, In-Phase, Out-phase image or None"""

        on_disk = self.on_disk()
        if on_disk: self.read()
        flagWater = False
        flagFat = False
        flagInPhase = False
        flagOutPhase = False
        if hasattr(self.ds, 'ImageType'):
            type = set(self.ds.ImageType)
            if set(['W', 'WATER']).intersection(type):
                flagWater = True
            elif set(['F', 'FAT']).intersection(type):# or ('B0' in dataset.ImageType) or ('FIELD_MAP' in dataset.ImageType):
                flagFat = True
            elif set(['IP', 'IN_PHASE']).intersection(type):
                flagInPhase = True
            elif set(['OP', 'OUT_PHASE']).intersection(type):
                flagOutPhase = True
        if on_disk: self.clear()
        return flagWater, flagFat, flagInPhase, flagOutPhase





class StatusBar():
    """
    Class with the same interface as StatusBar for use outside weasel.
    """

    def __init__(self):
        self._message = ''

    def hide(self):
        pass

    def message(self, message):
        self._message = message
        print(message)
        
    def progress(self, value, maximum, message=None):
        if message is not None:
            self.message(message)
        perc = str(100*value/maximum)
        print(self._message + ' (' + perc + ' %)')
        

class Dialog():
    """Class with the same interface as widgets.Dialog for commandline operation"""

    def information(self, message="Message in the box", title=""):
        """
        Information message. 
        """
        print("INFORMATION")
        if title != "": print(title)
        print(message)

    def warning(self, message="Message in the box", title=""):
        """
        Warning message. 
        """
        print("WARNING!")
        if title != "": print(title)
        print(message)

    def error(self, message="Message in the box", title=""):
        """
        Error message.
        """
        print("ERROR!")
        if title != "": print(title)
        print(message)

    def directory(self, message='Please select a folder', datafolder=None):
        """
        Select a directory.
        """
        pass

    def question(self, message="Do you wish to proceed?", title="Question", cancel=False):
        """
        Displays a question window in the User Interface.
        
        The user has to click either "OK" or "Cancel" in order to continue using the interface.
        Returns False if reply is "Cancel" and True if reply is "OK".
        """
        if cancel:
            instructions = "y = yes, n = no, c = cancel"
            options = ['y', 'n', 'c']
        else:
            instructions = "y = yes, n = no, c = cancel"
            options = ['y', 'n']
        print(title)
        print(message)
        print(instructions)
        answer = input()
        while answer not in options:
            print("Sorry, I can't interpret that answer")
            print(message)
            print(instructions)
            answer = input()
        if answer == 'y': return "Yes"
        if answer == 'n': return "No"
        if answer == 'c': return "Cancel"

    def file_to_open(self, title='Open file..', initial_folder=None, extension="All files (*.*)", datafolder=None):
        """
        Select a file to read.
        """
        pass

    def file_to_save(self, title='Save as ...', directory=None, filter="All files (*.*)", datafolder=None):
        """
        Select a filename to save.
        """
        pass

    def input(self, *fields, title="User input window"):
        """
        Collect user input of various types.
        """
        pass



"""Defines the central `Folder` class for reading and writing to DICOM folders.

`Folder()` is the central class of `dbdicom` and the first point of call 
for opening, closing and manipulating a folder with DICOM files. 

    # Example: Get a 3D numpy array from the first series in a folder.

    from dbdicom import Folder

    folder = Folder('C:\\Users\\MyName\\MyData\\DICOMtestData')
    array = folder.open().series(0).array()
"""

__all__ = ['Folder']

import os
import pydicom
import pandas as pd

#from dbdicom import dicm, utilities
from . import dicm, utilities
from .message import StatusBar, Dialog
from .classes.database import Database

class Folder(Database): 
    # This really needs to be Database() with Folder as an attribute. 
    """Programming interface for reading and writing a DICOM folder."""

    # The column labels of the dataframe as required by dbdicom
    required = [    
        'PatientID', 'StudyInstanceUID', 'SeriesInstanceUID', 'SOPInstanceUID', 
        'SOPClassUID','NumberOfFrames', 
        'PatientName', 
        'StudyDescription', 'StudyDate', 
        'SeriesDescription', 'SeriesNumber',
        'InstanceNumber', 'AcquisitionTime', 
    ]

    def __init__(self, path=None, attributes=None, message='Opening folder..', status=StatusBar(), dialog=Dialog()):
        """Initialise the folder with a path and objects to message to the user.
        
        When used inside a GUI, status and dialog should be instances of the status bar and 
        dialog class defined in `weasel`.
        """
        # A list of optional dicom attributes that are not used by dbdicom 
        # but can be included for faster access in other applications. 
        # The list can be changed or extended by applications.
        self.__dict__['attributes'] = ['SliceLocation']
        self.__dict__['dataframe'] = pd.DataFrame([]*len(self._columns), columns=self._columns)            
        self.__dict__['path'] = path
        self.__dict__['status'] = status
        self.__dict__['dialog'] = dialog
        self.__dict__['dicm'] = dicm

        # These are anomalies - folder should not inherit DataBase
        self.__dict__['UID'] = []       
        self.__dict__['folder'] = self
        self.__dict__['ds'] = None

        self.set_attributes(attributes, scan=False)
        self.open(message=message)


    @property
    def _columns(self):

        return self.required + self.__dict__['attributes']

    def set_attributes(self, attributes, scan=True):
        """DICOM attributes that are NOT used by dbdicom.
        
        Can be set by applications for fast access to key DICOM attributes such
        as those used for sorting series data.

        Args:
            attributes: list of DICOM attributes.
        """
        if attributes is None:
            return
        # Make sure all columns are unique
        attr = []
        for a in attributes:
            if a not in self.required:
                attr.append(a)
        self.__dict__['attributes'] = attr
        if scan: self.scan()

    def open(self, path=None, message='Opening folder..', unzip=False):
        """Opens a DICOM folder for read and write.
        
        Reads the contents of the folder and summarises all DICOM files
        in a dataframe for faster access next time. The dataframe is saved 
        as a csv file when the folder is closed with `folder.close()`. 
        All non-DICOM files in the folder are ignored.
        
        Args:
            path: The full path to the directory that is to be opened.

        Returns:
            The folder instance. This allows using the open() method in a 
            piping notation as in `inst = folder.open().instances()`.
        """
        if path is not None: 
            self.__dict__['path'] = path
        if self.path is None:
            message = "please set a path before opening."
            message += "\n Use folder.open(path_to_dicom_folder)."
            self.dialog.information(message)
            return self
        if os.path.exists(self._csv):
            self.status.message("Reading register..")
            self._read_csv()
            # If the saved register does not have all required attributes
            # then scan the folder again and create a new register
            labels = self._columns + ['removed','created']
            if labels != list(self.dataframe.columns):
#            for attribute in self.__dict__['attributes']:
#                if attribute not in self.dataframe:
                self.scan(message=message)
                self.status.hide()
                return self
            self.status.hide()
        else:
            self.scan(message=message, unzip=unzip)
        return self

    def scan(self, message='Scanning..', unzip=False):
        """
        Reads all files in the folder and summarises key attributes in a table for faster access.
        """
        if unzip:
            self.status.message('Extracting compressed folders..')
            utilities._unzip_files(self.path, self.status)
        self.status.message('Finding all files..')
        files = [item.path for item in utilities.scan_tree(self.path) if item.is_file()]
        self.__dict__['dataframe'] = utilities.dataframe(self.path, files, self._columns, self.status, message=message)
        self.dataframe['removed'] = [False] * self.dataframe.shape[0]
        self.dataframe['created'] = [False] * self.dataframe.shape[0]
        self._multiframe_to_singleframe()
        self._write_csv()
        return self

    def close(self):
        """Close an open folder.
        
        This method checks if the changes have been saved and prompts the 
        user to save or restore them if not. The user also has the option 
        to cancel closing, in which case the function does nothing. 
        
        `close()` also resets the dataframe and path to default values.

        Returns: 
            True if the user has agreed to close the folder 
            (possible after save or restore) and False if the user 
            has cancelled the closing of the folder.
        """
        if not self.is_open(): 
            return True
        if not self.is_saved():
            reply = self.dialog.question( 
                title = 'Closing DICOM folder', 
                message = 'Save changes before closing?',
                cancel = True, 
            )
            if reply == "Cancel": 
                return False
            if reply == "Yes":
                self.save()
            elif reply == "No":
                self.restore()

        self._write_csv()
        self.__dict__['dataframe'] = pd.DataFrame([]*len(self._columns), columns=self._columns)            
        self.__dict__['path'] = None

        return True

    def save(self):

        self.status.message("Saving..")
        if self.is_saved():
            self._write_csv()
        else:
            super().save()
        self.status.hide()

    def is_saved(self):
        """Check if the folder is saved.
        
        Returns: 
            True if the folder is saved and False otherwise.
        """
        if self.dataframe.removed.any(): return False
        if self.dataframe.created.any(): return False
        return True

    def is_open(self):
        """Check if the folder is currently open.
        
        Returns: 
            True if the folder is open and False otherwise.
        """
        return self.path is not None

    def sortby(self, sortby):
        
        self.dataframe.sort_values(sortby, inplace=True)
        return self

    def object(self, row, generation=4):
        """Create a new dicom object from a row in the dataframe.
        
        Args:
            row: a row in the dataframe as a series
            generation: determines whether the object returned is 
                at level database (generation=0), patient (generation=1),
                study (generation=2), series (generation=3) 
                or instance (generation=4, default).
        Returns:
            An instance of one of the dicom classes defined in wedicom.classes.
        """
        return dicm.object(self, row, generation)

    def instance(self, file):
        """Create an instance from a filepath"""

        row = self.dataframe.loc[file]
        return self.object(row) 

    def print(self):
        """Prints a summary of the project folder to the terminal."""
        
        print(' ')
        print('---------- DICOM FOLDER --------------')
        print('FOLDER: ' + self.path)
        for i, patient in enumerate(self.children()):
            print(' ')
            print('    PATIENT [' + str(i) + ']: ' + patient.label())
            print(' ')
            for j, study in enumerate(patient.children()):
                print('        STUDY [' + str(j) + ']: ' + study.label())
                print(' ')
                for k, series in enumerate(study.children()):
                    print('            SERIES [' + str(k) + ']: ' + series.label())
                    print('                Nr of instances: ' + str(len(series.children()))) 

    def _write_csv(self):
        """ Writes the dataFrame as a .csv file"""
        file = self._csv
        self.dataframe.to_csv(file)

    def _read_csv(self):
        """Reads the dataFrame from a .csv file """  
        file = self._csv
        self.__dict__['dataframe'] = pd.read_csv(file, index_col=0)

    @property
    def _csv(self):
        """ Returns the file path of the .csv file"""
        filename = os.path.basename(os.path.normpath(self.path)) + ".csv"
        return os.path.join(self.path, filename) 

    def label(self, row, type):

        if type == 'Patient':
            name = row.PatientName
            id = row.PatientID
            label = str(name)
            label += ' [' + str(id) + ']'
            return type + " - {}".format(label)
        if type == 'Study':
            descr = row.StudyDescription
            date = row.StudyDate
            label = str(descr)
            label += ' [' + str(date) + ']'
            return type + " - {}".format(label)
        if type == 'Series':
            descr = row.SeriesDescription
            nr = row.SeriesNumber
            label = '[' + str(nr).zfill(3) + '] ' 
            label += str(descr)
            return type + " - {}".format(label)
        if type == 'Instance':
            nr = row.InstanceNumber
            label = str(nr).zfill(6)
            return type + " - {}".format(label)

    def new_uid(self, n=1):
        
        if n == 1:
            return pydicom.uid.generate_uid()
        uid = []
        for _ in range(n):
            uid.append(pydicom.uid.generate_uid())
        return uid

    def new_file(self):

        # Generate a new filename
        path = os.path.join(self.path, "dbdicom")
        if not os.path.isdir(path): os.mkdir(path)
        #file = os.path.join(path, self.new_uid() + '.dcm') 
        file = os.path.join("dbdicom", self.new_uid() + '.dcm') 
        return file

    def _append(self, ds):
        """Append a new row to the dataframe from a pydicom dataset.
        
        Args:
            ds: and instance of pydicom dataset
        """

        # Generate a new filename
        file = self.new_file()
#        path = os.path.join(self.path, "Weasel")
#        if not os.path.isdir(path): os.mkdir(path)
#        file = os.path.join(path, pydicom.uid.generate_uid() + '.dcm') 

        row = pd.DataFrame([]*len(self._columns), index=[file], columns=self._columns)
        row['removed'] = False
        row['created'] = True
        self.__dict__['dataframe'] = pd.concat([self.dataframe, row]) 
        
        # REPLACED BY CONCAT on 03 june 2022
        # labels = self._columns + ['removed','created']
        #row = pd.Series(data=['']*len(labels), index=labels, name=file)
        #row['removed'] = False
        #row['created'] = True
        #self.__dict__['dataframe'] = self.dataframe.append(row) # REPLACE BY CONCAT

        self._update(file, ds)

    def _update(self, file, ds):
        """Update the values of a dataframe row.
        
        Args:
            file: filepath (or index) of the row to be updated
            ds: instance of a pydicom dataset.
        """
        for tag in self._columns:
            if tag in ds:
                value = utilities._read_tags(ds, tag)
                self.dataframe.loc[file, tag] = value
                #self.dataframe.loc[file, tag] = ds[tag].value

    def _multiframe_to_singleframe(self):
        """Converts all multiframe files in the folder into single-frame files.
        
        Reads all the multi-frame files in the folder,
        converts them to singleframe files, and delete the original multiframe file.
        """

        singleframe = self.dataframe.NumberOfFrames.isnull() 
        multiframe = singleframe == False
        nr_multiframe = multiframe.sum()
        if nr_multiframe != 0: 
            cnt=0
            for filepath, instance in self.dataframe[multiframe].iterrows():
                cnt+=1
                self.status.message("Converting multiframe file " + filepath)
                self.status.progress(cnt, nr_multiframe)
                #
                # Create these in the Weasel folder, not in the original folder.
                #
                singleframe_files = utilities.split_multiframe(filepath, str(instance.SeriesInstanceUID))
                if singleframe_files != []:                    
                    # add the single frame files to the dataframe
                    df = utilities.dataframe(self.path, singleframe_files, self._columns)
                    df['removed'] = [False] * df.shape[0]
                    df['created'] = [False] * df.shape[0]
                    self.__dict__['dataframe'] = pd.concat([self.dataframe, df])
                    # delete the original multiframe 
                    # CHANGE THIS - mark for removal but don't delete
                    # show enhanced DICOM in Tree but set the class to block display etc.
                    os.remove(filepath)
                    self.dataframe.drop(index=filepath, inplace=True)
                self.status.hide()


# setup.py
# Dummy file
# This file is there to fix a limitation in setuptools
# https://realpython.com/pypi-publish-python-package/#prepare-your-package-for-publication

from setuptools import setup

setup()


from pydicom.dataset import Dataset
import numpy as np
import datetime


  
class ParametricMap(object):
    def selectParametricMap(self, dicom, imageArray, argument):
        methodName = argument
        method = getattr(self, methodName, lambda: "No valid Parametric Map chosen")
        return method(dicom, imageArray)

    def RGB(self, dicom, imageArray):
        dicom.PhotometricInterpretation = 'RGB'
        dicom.SamplesPerPixel = 3
        dicom.BitsAllocated = 8
        dicom.BitsStored = 8
        dicom.HighBit = 7
        dicom.add_new(0x00280006, 'US', 0) # Planar Configuration
        dicom.RescaleSlope = 1
        dicom.RescaleIntercept = 0
        pixelArray = imageArray.astype(np.uint8) # Should we multiply by 255?
        dicom.WindowCenter = int((np.amax(imageArray) - np.amin(imageArray)) / 2)
        dicom.WindowWidth = np.absolute(int(np.amax(imageArray) - np.amin(imageArray)))
        dicom.PixelData = pixelArray.tobytes()
        return

    def ADC(self, dicom, imageArray):
        # The commented parts are to apply when we decide to include Parametric Map IOD. No readers can deal with this yet
        # dicom.SOPClassUID = '1.2.840.10008.5.1.4.1.1.67'
        dicom.SeriesDescription = "Apparent Diffusion Coefficient (um2/s)"
        dicom.Modality = "RWV"
        dicom.FrameLaterality = "U"
        dicom.DerivedPixelContrast = "ADC"
        dicom.BitsAllocated = 32
        dicom.PixelRepresentation = 1
        dicom.PhotometricInterpretation = "MONOCHROME2"
        dicom.PixelAspectRatio = ["1", "1"] # Need to have a better look at this
        dicom.RescaleSlope = 1
        dicom.RescaleIntercept = 0
        # Rotate the image back to the original orientation
        imageArray = np.transpose(imageArray)
        dicom.Rows = np.shape(imageArray)[-2]
        dicom.Columns = np.shape(imageArray)[-1]
        dicom.WindowCenter = int((np.amax(imageArray) - np.amin(imageArray)) / 2)
        dicom.WindowWidth = np.absolute(int(np.amax(imageArray) - np.amin(imageArray)))
        dicom.FloatPixelData = bytes(imageArray.astype(np.float32).flatten())
        del dicom.PixelData, dicom.BitsStored, dicom.HighBit

        dicom.RealWorldValueMappingSequence = [Dataset(), Dataset(), Dataset(), Dataset()]
        dicom.RealWorldValueMappingSequence[0].QuantityDefinitionSequence = [Dataset(), Dataset()]
        dicom.RealWorldValueMappingSequence[0].QuantityDefinitionSequence[0].ValueType = "CODE"
        dicom.RealWorldValueMappingSequence[0].QuantityDefinitionSequence[1].ConceptCodeSequence = [Dataset(), Dataset(), Dataset()]
        dicom.RealWorldValueMappingSequence[0].QuantityDefinitionSequence[1].ConceptCodeSequence[0].CodeValue = "113041"
        dicom.RealWorldValueMappingSequence[0].QuantityDefinitionSequence[1].ConceptCodeSequence[1].CodingSchemeDesignator = "DCM"
        dicom.RealWorldValueMappingSequence[0].QuantityDefinitionSequence[1].ConceptCodeSequence[2].CodeMeaning = "Apparent Diffusion Coefficient"
        dicom.RealWorldValueMappingSequence[1].MeasurementUnitsCodeSequence = [Dataset(), Dataset(), Dataset()]
        dicom.RealWorldValueMappingSequence[1].MeasurementUnitsCodeSequence[0].CodeValue = "um2/s"
        dicom.RealWorldValueMappingSequence[1].MeasurementUnitsCodeSequence[1].CodingSchemeDesignator = "UCUM"
        dicom.RealWorldValueMappingSequence[1].MeasurementUnitsCodeSequence[2].CodeMeaning = "um2/s"
        dicom.RealWorldValueMappingSequence[2].RealWorldValueSlope = 1
        
        anatomyString = dicom.BodyPartExamined
        saveAnatomicalInfo(anatomyString, dicom.RealWorldValueMappingSequence[3])

        return

    def T2Star(self, dicom, imageArray):
        dicom.PixelSpacing = [3, 3] # find a mechanism to pass reconstruct pixel here
        return

    def SEG(self, dicom, imageArray):
        #dicom.SOPClassUID = '1.2.840.10008.5.1.4.1.1.66.4' # WILL NOT BE USED HERE - This is for PACS. There will be another one for DICOM Standard
        # The commented parts are to apply when we decide to include SEG IOD. No readers can deal with this yet
        dicom.BitsAllocated = 8 # According to Federov DICOM Standard this should be 1-bit
        dicom.BitsStored = 8
        dicom.HighBit = 7
        #dicom.SmallestImagePixelValue = 0
        #dicom.LargestImagePixelValue = int(np.amax(imageArray)) # max 255
        dicom.add_new('0x00280106', 'US', 0) # Minimum
        dicom.add_new('0x00280107', 'US', int(np.amax(imageArray))) # Maximum
        dicom.PixelRepresentation = 0
        dicom.SamplesPerPixel = 1
        #dicom.WindowCenter = 0.5
        #dicom.WindowWidth = 1.1
        dicom.add_new('0x00281050', 'DS', 0.5) # WindowCenter
        dicom.add_new('0x00281051', 'DS', 1.1) # WindowWidth
        #dicom.RescaleIntercept = 0
        #dicom.RescaleSlope = 1
        dicom.add_new('0x00281052', 'DS', 0) # RescaleIntercept
        dicom.add_new('0x00281053', 'DS', 1) # RescaleSlope
        dicom.LossyImageCompression = '00'
        pixelArray = np.transpose(imageArray.astype(np.uint8)) # Should we multiply by 255?
        dicom.PixelData = pixelArray.tobytes()

        dicom.Modality = 'SEG'
        dicom.SegmentationType = 'FRACTIONAL'
        dicom.MaximumFractionalValue = int(np.amax(imageArray)) # max 255
        dicom.SegmentationFractionalType = 'OCCUPANCY'

        # Segment Labels
        if hasattr(dicom, "ImageComments"):
            dicom.ContentDescription = dicom.ImageComments.split('_')[-1] # 'Image segmentation'
            segment_numbers = np.unique(pixelArray)
            segment_dictionary = dict(list(enumerate(segment_numbers)))
            segment_label = dicom.ImageComments.split('_')[-1]
            segment_dictionary[0] = 'Background'
            segment_dictionary[1] = segment_label
            for key in segment_dictionary:
                dicom.SegmentSequence = [Dataset(), Dataset(), Dataset(), Dataset(), Dataset(), Dataset()]
                dicom.SegmentSequence[0].SegmentAlgorithmType = 'MANUAL'
                dicom.SegmentSequence[1].SegmentNumber = key
                dicom.SegmentSequence[2].SegmentDescription = str(segment_dictionary[key])
                dicom.SegmentSequence[3].SegmentLabel = str(segment_dictionary[key])
                dicom.SegmentSequence[4].SegmentAlgorithmName = "Weasel"
                if hasattr(dicom, "BodyPartExamined"):
                    anatomyString = dicom.BodyPartExamined
                    saveAnatomicalInfo(anatomyString, dicom.SegmentSequence[5])
        else:
            dicom.ContentDescription = "Mask with no label"

        return

    def Registration(self, dicom, imageArray):
        dicom.Modality = "REG"
        return

    def Signal(self, dicom, imageArray):
        dicom.Modality = "RWV"
        dicom.DerivedPixelContrast = "GraphPlot"
        dicom.PhotometricInterpretation = "MONOCHROME2"
        dicom.RescaleSlope = 1
        dicom.RescaleIntercept = 0
        imageArray = np.transpose(imageArray.astype(np.float32))
        center = (np.amax(imageArray) + np.amin(imageArray)) / 2
        width = np.amax(imageArray) - np.amin(imageArray)
        dicom.add_new('0x00281050', 'DS', center)
        dicom.add_new('0x00281051', 'DS', width)
        dicom.BitsAllocated = 32
        dicom.Rows = np.shape(imageArray)[0]
        dicom.Columns = np.shape(imageArray)[1]
        dicom.FloatPixelData = bytes(imageArray.flatten())
        del dicom.PixelData, dicom.BitsStored, dicom.HighBit
        return

# Could insert a method regarding ROI colours, like in ITK-SNAP???
def saveAnatomicalInfo(anatomyString, dicom):
    try:
        # FOR NOW, THE PRIORITY WILL BE ON KIDNEY
        if "KIDNEY" or "ABDOMEN" in anatomyString.upper():
            dicom.AnatomicRegionSequence = [Dataset(), Dataset(), Dataset()]
            dicom.AnatomicRegionSequence[0].CodeValue = "T-71000"
            dicom.AnatomicRegionSequence[1].CodingSchemeDesignator = "SRT"
            dicom.AnatomicRegionSequence[2].CodeMeaning = "Kidney"
        elif "LIVER" in anatomyString.upper():
            dicom.AnatomicRegionSequence = [Dataset(), Dataset(), Dataset()]
            dicom.AnatomicRegionSequence[0].CodeValue = "T-62000"
            dicom.AnatomicRegionSequence[1].CodingSchemeDesignator = "SRT"
            dicom.AnatomicRegionSequence[2].CodeMeaning = "Liver"
        elif "PROSTATE" in anatomyString.upper():
            dicom.AnatomicRegionSequence = [Dataset(), Dataset(), Dataset()]
            dicom.AnatomicRegionSequence[0].CodeValue = "T-9200B"
            dicom.AnatomicRegionSequence[1].CodingSchemeDesignator = "SRT"
            dicom.AnatomicRegionSequence[2].CodeMeaning = "Prostate"      
        elif "BODY" in anatomyString.upper():
            dicom.AnatomicRegionSequence = [Dataset(), Dataset(), Dataset()]
            dicom.AnatomicRegionSequence[0].CodeValue = "P5-0905E"
            dicom.AnatomicRegionSequence[1].CodingSchemeDesignator = "LN"
            dicom.AnatomicRegionSequence[2].CodeMeaning = "MRI whole body"
    except:
        pass
    return

def editDicom(newDicom, imageArray, parametricMap):

    callCase = ParametricMap()
    callCase.selectParametricMap(newDicom, imageArray, parametricMap)

    dt = datetime.datetime.now()
    timeStr = dt.strftime('%H%M%S')  # long format with micro seconds
    newDicom.PerformedProcedureStepStartDate = dt.strftime('%Y%m%d')
    newDicom.PerformedProcedureStepStartTime = timeStr
    newDicom.PerformedProcedureStepDescription = "Post-processing application"

    return newDicom

    # Series, Instance and Class for Reference
    #newDicom.ReferencedSeriesSequence = [Dataset(), Dataset()]
    #newDicom.ReferencedSeriesSequence[0].SeriesInstanceUID = dicom_data.SeriesInstanceUID
    #newDicom.ReferencedSeriesSequence[1].ReferencedInstanceSequence = [Dataset(), Dataset()]
    #newDicom.ReferencedSeriesSequence[1].ReferencedInstanceSequence[0].ReferencedSOPClassUID = dicom_data.SOPClassUID
    #newDicom.ReferencedSeriesSequence[1].ReferencedInstanceSequence[1].ReferencedSOPInstanceUID = dicom_data.SOPInstanceUID

# rwv_sequence = Sequence()
        # dicom.RealWorldValueMappingSequence = rwv_sequence
        # rwv_slope = Dataset()
        # rwv_slope.RealWorldValueSlope = 1
        # rwv_sequence.append(rwv_slope)

        # quantity_def = Dataset()
        # quantity_def_sequence = Sequence()
        # quantity_def.QuantityDefinitionSequence = quantity_def_sequence
        # value_type = Dataset()
        # value_type.ValueType = "CODE"
        # quantity_def_sequence.append(value_type)
        # concept_code = Dataset()
        # concept_code_sequence = Sequence()
        # concept_code.ConceptCodeSequence = concept_code_sequence
        # code_code = Dataset()
        # code_code.CodeValue = "113041"
        # code_code.CodingSchemeDesignator = "DCM"
        # code_code.CodeMeaning = "Apparent Diffusion Coefficient"
        # concept_code_sequence.append(code_code)
        # rwv_sequence.append(quantity_def)

        # measure_units = Dataset()
        # measure_units_sequence = Sequence()
        # measure_units.MeasurementUnitsCodeSequence = measure_units_sequence
        # measure_code = Dataset()
        # measure_code.CodeValue = "um2/s"
        # measure_code.CodingSchemeDesignator = "UCUM"
        # measure_code.CodeMeaning = "um2/s"
        # measure_units_sequence.append(measure_code)
        # rwv_sequence.append(measure_units)